from datetime import timedelta
from decimal import Decimal

from rest_framework import serializers
from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers
from wbfdm.serializers import InvestableInstrumentRepresentationSerializer

from wbportfolio.models import PortfolioRole, Trade, TradeProposal
from wbportfolio.models.transactions.claim import Claim
from wbportfolio.serializers.custodians import CustodianRepresentationSerializer
from wbportfolio.serializers.registers import RegisterRepresentationSerializer

from .transactions import (
    TransactionModelSerializer,
    TransactionRepresentationSerializer,
)


class CopyClaimRepresentationSerializer(wb_serializers.RepresentationSerializer):
    claimant = wb_serializers.StringRelatedField()
    product = wb_serializers.StringRelatedField()
    account = wb_serializers.StringRelatedField()
    _detail = wb_serializers.HyperlinkField(reverse_name="wbportfolio:claim-detail")

    class Meta:
        model = Claim
        fields = (
            "id",
            "shares",
            "claimant",
            "bank",
            "product",
            "account",
            "_detail",
        )


class TradeProposalRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbportfolio:tradeproposal-detail")

    class Meta:
        model = TradeProposal
        fields = ("id", "trade_date", "status", "_detail")


class TradeRepresentationSerializer(TransactionRepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbportfolio:trade-detail")
    diff_shares = wb_serializers.DecimalField(max_digits=15, decimal_places=4)

    class Meta:
        model = Trade
        fields = ("claimed_shares", "diff_shares", "bank", "shares") + TransactionRepresentationSerializer.Meta.fields


class TradeClaimRepresentationSerializer(TradeRepresentationSerializer):
    def get_filter_params(self, request):
        claim_id = request.parser_context["view"].kwargs.get("pk", None)
        arg = {"marked_for_deletion": False}
        if claim_id:
            claim = Claim.objects.get(id=claim_id)
            arg.update(
                {
                    "underlying_instrument": claim.product.id,
                    "transaction_date__gte": claim.date - timedelta(days=Trade.TRADE_WINDOW_INTERVAL),
                    "transaction_date__lte": claim.date + timedelta(days=Trade.TRADE_WINDOW_INTERVAL),
                }
            )
            if claim.shares:
                if claim.shares > 0:
                    arg["shares__gte"] = claim.shares
                else:
                    arg["shares__lte"] = claim.shares
            return arg
        return {}


class TradeModelSerializer(TransactionModelSerializer):
    marked_for_deletion = wb_serializers.BooleanField(required=False, read_only=True)
    marked_as_internal = wb_serializers.BooleanField()

    approved_claimed_shares = wb_serializers.DecimalField(
        max_digits=10, decimal_places=2, required=False, read_only=True, default=Decimal(0.0)
    )
    pending_claimed_shares = wb_serializers.DecimalField(
        max_digits=10, decimal_places=2, required=False, read_only=True, default=Decimal(0.0)
    )
    completely_claimed = wb_serializers.BooleanField(required=False, read_only=True, default=Decimal(0.0))
    completely_claimed_if_approved = wb_serializers.BooleanField(required=False, read_only=True, default=Decimal(0.0))

    # We commented it because people needs to filter with lookup equals onto these fields
    # shares = wb_serializers.DecimalField(max_digits=16, decimal_places=2)
    # price = wb_serializers.DecimalField(max_digits=16, decimal_places=2)

    _register = RegisterRepresentationSerializer(source="register")
    claims = wb_serializers.PrimaryKeyRelatedField(queryset=Claim.objects.all(), label="Claims", many=True)
    _claims = CopyClaimRepresentationSerializer(source="claims", many=True)
    _custodian = CustodianRepresentationSerializer(source="custodian")
    _internal_trade = TradeRepresentationSerializer(
        source="internal_trade",
        filter_params={"only_internal_trade": True},
        optional_get_parameters={"underlying_instrument": "underlying_instrument", "transaction_date": "pivot_date"},
        depends_on=[{"field": "marked_as_internal", "options": {"activates_on": [True]}}],
    )

    @wb_serializers.register_only_instance_resource()
    def import_source_file_url(self, instance, request, user, *args, **kwargs):
        if PortfolioRole.is_portfolio_manager(user.profile, instance.underlying_instrument):
            if instance.import_source:
                return {"import_source": instance.import_source.file.url}
        return {}

    @wb_serializers.register_resource()
    def claims_list(self, instance, request, user):
        if instance.is_claimable:
            return {
                "claims": reverse(
                    "wbportfolio:trade-claim-list",
                    args=[instance.id],
                    request=request,
                )
            }
        return {}

    class Meta:
        model = Trade
        decorators = TransactionModelSerializer.Meta.decorators
        read_only_fields = (
            "id",
            "shares",
            "price",
            "transaction_subtype",
            "approved_claimed_shares",
            "pending_claimed_shares",
            "claimed_shares",
            "completely_claimed",
            "completely_claimed_if_approved",
            "register",
            "_register",
            "pending",
            "bank",
            "marked_for_deletion",
            "_claims",
            "claims",
            "transaction_type",
            "transaction_url_type",
            "transaction_underlying_type",
            "portfolio",
            "_portfolio",
            "underlying_instrument",
            "_underlying_instrument",
            "transaction_date",
            "book_date",
            "value_date",
            "currency",
            "_currency",
            "currency_fx_rate",
            "total_value",
            "total_value_fx_portfolio",
            "total_value_gross",
            "total_value_gross_fx_portfolio",
            "external_id",
            "total_value_usd",
            "total_value_gross_usd",
            "_additional_resources",
        )

        fields = (
            "id",
            "shares",
            "price",
            "transaction_subtype",
            "approved_claimed_shares",
            "pending_claimed_shares",
            "claimed_shares",
            "completely_claimed",
            "completely_claimed_if_approved",
            "register",
            "_register",
            "pending",
            "bank",
            "marked_for_deletion",
            "_claims",
            "claims",
            "transaction_type",
            "transaction_url_type",
            "transaction_underlying_type",
            "portfolio",
            "_portfolio",
            "underlying_instrument",
            "_underlying_instrument",
            "transaction_date",
            "book_date",
            "value_date",
            "currency",
            "_currency",
            "currency_fx_rate",
            "total_value",
            "total_value_fx_portfolio",
            "total_value_gross",
            "total_value_gross_fx_portfolio",
            "external_id",
            "comment",
            "total_value_usd",
            "total_value_gross_usd",
            "_custodian",
            "custodian",
            "marked_as_internal",
            "internal_trade",
            "_internal_trade",
            "_additional_resources",
        )


class TradeTradeProposalModelSerializer(TradeModelSerializer):
    underlying_instrument_isin = wb_serializers.CharField(read_only=True)
    underlying_instrument_ticker = wb_serializers.CharField(read_only=True)
    underlying_instrument_refinitiv_identifier_code = wb_serializers.CharField(read_only=True)
    underlying_instrument_instrument_type = wb_serializers.CharField(read_only=True)
    _underlying_instrument = InvestableInstrumentRepresentationSerializer(
        source="underlying_instrument", label_key="name"
    )

    status = wb_serializers.ChoiceField(default=Trade.Status.DRAFT, choices=Trade.Status.choices)
    target_weight = wb_serializers.DecimalField(max_digits=16, decimal_places=6, required=False, default=0)
    effective_weight = wb_serializers.DecimalField(read_only=True, max_digits=16, decimal_places=6, default=0)
    effective_shares = wb_serializers.DecimalField(read_only=True, max_digits=16, decimal_places=6, default=0)
    target_shares = wb_serializers.DecimalField(read_only=True, max_digits=16, decimal_places=6, default=0)

    def validate(self, data):
        if self.instance and "underlying_instrument" in data:
            raise serializers.ValidationError(
                {
                    "underlying_instrument": "You cannot modify the underlying instrument other than creating a new entry"
                }
            )
        effective_weight = self.instance._effective_weight if self.instance else Decimal(0.0)
        if (target_weight := data.pop("target_weight", None)) is not None:
            data["weighting"] = target_weight - effective_weight
        if data["weighting"] >= 0:
            data["transaction_subtype"] = "BUY"
        else:
            data["transaction_subtype"] = "SELL"
        return data

    class Meta:
        model = Trade
        percent_fields = ["effective_weight", "target_weight", "weighting"]
        read_only_fields = (
            "transaction_subtype",
            "shares",
            # "underlying_instrument",
            # "_underlying_instrument",
            "shares",
            "effective_shares",
            "target_shares",
        )
        fields = (
            "id",
            "shares",
            "underlying_instrument_isin",
            "underlying_instrument_ticker",
            "underlying_instrument_refinitiv_identifier_code",
            "underlying_instrument_instrument_type",
            "underlying_instrument",
            "_underlying_instrument",
            "transaction_subtype",
            "status",
            "comment",
            "effective_weight",
            "target_weight",
            "weighting",
            "trade_proposal",
            "order",
            "effective_shares",
            "target_shares",
        )


class ReadOnlyTradeTradeProposalModelSerializer(TradeTradeProposalModelSerializer):
    class Meta(TradeTradeProposalModelSerializer.Meta):
        read_only_fields = TradeTradeProposalModelSerializer.Meta.fields



import xlrd

from wbportfolio.models import Product


def get_isin(sheet):
    for row in sheet.get_rows():
        if row[0].value == "ISIN":
            return row[1].value


def parse(import_source):
    book = xlrd.open_workbook(file_contents=import_source.file.read())

    product = Product.objects.get(isin=get_isin(book.sheet_by_name("Certificate")))
    equity_sheet = book.sheet_by_name("Report - details")

    # date_column = None
    # net_value_column = None
    # shares_column = None

    data = list()

    first_row = None
    last_row = None

    for i, row in enumerate(equity_sheet.get_rows()):
        if row[0].value == "Date":
            first_row = i

        if first_row is not None and row[0].value == "":
            last_row = i

        if first_row is not None and last_row is not None:
            valuation_date = xlrd.xldate.xldate_as_datetime(equity_sheet.cell_value(first_row + 1, 0), 0).date()

            for _i in range(first_row + 3, last_row):
                initial_shares = equity_sheet.cell_value(_i, 2)
                isin = equity_sheet.cell_value(_i, 3)
                equity_sheet.cell_value(_i, 4)
                currency_key = equity_sheet.cell_value(_i, 5)
                initial_currency_fx_rate = equity_sheet.cell_value(_i, 6)
                initial_price = equity_sheet.cell_value(_i, 10)

                ticker = Product.objects.get(isin=isin).ticker

                data.append(
                    {
                        "underlying_quote": {"instrument_type": "product", "ticker": ticker},
                        "currency__key": currency_key,
                        "date": valuation_date.strftime("%Y-%m-%d"),
                        "asset_valuation_date": valuation_date.strftime("%Y-%m-%d"),
                        "initial_price": initial_price,
                        "initial_currency_fx_rate": initial_currency_fx_rate,
                        "initial_shares": initial_shares,
                        "portfolio": {"id": product.id, "instrument_type": "product"},
                    }
                )

            data.append(
                {
                    "underlying_quote": {
                        "instrument_type": "cash",
                        "ticker": "CASH",
                        "name": f"1 {product.currency.key} AMC",
                        "currency__key": product.currency.key,
                    },
                    "date": valuation_date.strftime("%Y-%m-%d"),
                    "asset_valuation_date": valuation_date.strftime("%Y-%m-%d"),
                    "initial_price": 1.0,
                    "initial_currency_fx_rate": 1.0,
                    "currency__key": product.currency.key,
                    "initial_shares": equity_sheet.cell_value(first_row + 1, 6),
                    "portfolio": {"id": product.id, "instrument_type": "product"},
                }
            )

            first_row = None
            last_row = None

    return {"data": data}



from typing import Optional

from wbcore.contrib.color.enums import WBColor
from wbcore.enums import Unit
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

from wbportfolio.filters.performances import PerformancePandasFilter


class PerformancePandasDisplayConfig(DisplayViewConfig):
    LEGENDS = [
        dp.Legend(
            items=[
                dp.LegendItem(icon=WBColor.GREEN_DARK.value, label="Strong Growth"),
                dp.LegendItem(icon=WBColor.GREEN.value, label="Medium Growth"),
                dp.LegendItem(icon=WBColor.GREEN_LIGHT.value, label="Slight Growth"),
                dp.LegendItem(icon=WBColor.RED_LIGHT.value, label="Slight Loss"),
                dp.LegendItem(icon=WBColor.RED.value, label="Medium Loss"),
                dp.LegendItem(icon=WBColor.RED_DARK.value, label="Strong Loss"),
            ]
        )
    ]
    FORMATTING = [
        dp.Formatting(
            column="perf_usd",
            formatting_rules=[
                dp.FormattingRule(
                    style={"backgroundColor": WBColor.GREEN_LIGHT.value},
                    condition=(">", 0),
                ),
                dp.FormattingRule(
                    style={"backgroundColor": WBColor.GREEN.value},
                    condition=(">", 0.01),
                ),
                dp.FormattingRule(
                    style={"backgroundColor": WBColor.GREEN_DARK.value},
                    condition=(">", 0.03),
                ),
                dp.FormattingRule(
                    style={"backgroundColor": WBColor.RED_LIGHT.value},
                    condition=("<", 0),
                ),
                dp.FormattingRule(style={"backgroundColor": WBColor.RED.value}, condition=("<", -0.01)),
                dp.FormattingRule(
                    style={"backgroundColor": WBColor.RED_DARK.value},
                    condition=("<", -0.03),
                ),
            ],
        ),
    ]

    def get_list_display(self) -> Optional[dp.ListDisplay]:
        n1_label = "N1"
        n2_label = "N2"
        if (
            self.request
            and (dates_str := self.request.GET.get("dates"))
            and (performance_by := self.request.GET.get("performance_by"))
        ):
            dates = dates_str.split(",")
            n1_label = f"{dates[0]} {PerformancePandasFilter.PerformanceBy[performance_by].label}"
            n2_label = f"{dates[1]} {PerformancePandasFilter.PerformanceBy[performance_by].label}"

        return dp.ListDisplay(
            fields=[
                dp.Field(key="computed_str", label="Title", width=Unit.PIXEL(300)),
                # dp.Field(key="isin", label="ISIN", width=Unit.PIXEL(150)),
                dp.Field(key="n1", label=n1_label),
                dp.Field(key="n2", label=n2_label),
                dp.Field(key="diff", label="Δ"),
                dp.Field(key="perf", label="Performance"),
                dp.Field(key="n1_usd", label=f"{n1_label} ($)"),
                dp.Field(key="n2_usd", label=f"{n2_label} ($)"),
                dp.Field(key="diff_usd", label="Δ ($)"),
                dp.Field(
                    key="perf_usd",
                    label="Performance ($)",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={
                                "fontWeight": "bold",
                            },
                        )
                    ],
                ),
            ],
            legends=self.LEGENDS,
            formatting=self.FORMATTING,
        )


class ProductPerformanceNetNewMoneyDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="computed_str", label="Title", width=Unit.PIXEL(300)),
                dp.Field(key="net_negative_money", label="Outflows", width=Unit.PIXEL(200)),
                dp.Field(key="net_negative_money_usd", label="Outflows ($)", width=Unit.PIXEL(200)),
                dp.Field(key="net_positive_money", label="Inflows", width=Unit.PIXEL(200)),
                dp.Field(key="net_positive_money_usd", label="Inflows ($)", width=Unit.PIXEL(200)),
                dp.Field(key="net_money", label="Net New Money", width=Unit.PIXEL(200)),
                dp.Field(key="net_money_usd", label="Net New Money ($)", width=Unit.PIXEL(200)),
            ],
            formatting=[
                dp.Formatting(
                    column="net_money",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREEN_LIGHT.value},
                            condition=(">=", 0),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.RED_LIGHT.value},
                            condition=("<", 0),
                        ),
                    ],
                ),
            ],
        )


class PerformanceComparisonDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="computed_str", label="Title", width=Unit.PIXEL(300)),
                dp.Field(key="benchmark_computed_str", label="Comparison benchmark", width=Unit.PIXEL(250))
                if self.view.benchmark_label
                else {},
                dp.Field(key="inception_date", label="Launch Date"),
                dp.Field(key="instrument_last_valuation_price", label="Last NAV (Price)"),
                dp.Field(key="last_valuation_date", label="Last NAV (Date)"),
                dp.Field(
                    key="perf_last_day",
                    label="Last Daily Perf.",
                    width=Unit.PIXEL(200),
                    formatting_rules=[
                        dp.FormattingRule(
                            style={
                                "color": WBColor.RED.value,
                                "fontWeight": "bold",
                            },
                            condition=("<", 0),
                        ),
                        dp.FormattingRule(
                            style={
                                "color": WBColor.GREEN_DARK.value,
                                "fontWeight": "bold",
                            },
                            condition=(">=", 0),
                        ),
                    ],
                ),
                dp.Field(
                    key="perf_month_to_date",
                    label="Month to Date",
                    width=Unit.PIXEL(200),
                    formatting_rules=[
                        dp.FormattingRule(
                            style={
                                "color": WBColor.RED.value,
                                "fontWeight": "bold",
                            },
                            condition=("<", 0),
                        ),
                        dp.FormattingRule(
                            style={
                                "color": WBColor.GREEN_DARK.value,
                                "fontWeight": "bold",
                            },
                            condition=(">=", 0),
                        ),
                    ],
                ),
                dp.Field(
                    key="perf_year_to_date",
                    label="Year to Date",
                    width=Unit.PIXEL(200),
                    formatting_rules=[
                        dp.FormattingRule(
                            style={
                                "color": WBColor.RED.value,
                                "fontWeight": "bold",
                            },
                            condition=("<", 0),
                        ),
                        dp.FormattingRule(
                            style={
                                "color": WBColor.GREEN_DARK.value,
                                "fontWeight": "bold",
                            },
                            condition=(">=", 0),
                        ),
                    ],
                ),
                dp.Field(
                    key="perf_inception",
                    label="Since Inception",
                    width=Unit.PIXEL(200),
                    formatting_rules=[
                        dp.FormattingRule(
                            style={
                                "color": WBColor.RED.value,
                                "fontWeight": "bold",
                            },
                            condition=("<", 0),
                        ),
                        dp.FormattingRule(
                            style={
                                "color": WBColor.GREEN_DARK.value,
                                "fontWeight": "bold",
                            },
                            condition=(">=", 0),
                        ),
                    ],
                ),
                dp.Field(
                    key="perf_between_dates",
                    label="Between Dates",
                    width=Unit.PIXEL(200),
                    formatting_rules=[
                        dp.FormattingRule(
                            style={
                                "color": WBColor.RED.value,
                                "fontWeight": "bold",
                            },
                            condition=("<", 0),
                        ),
                        dp.FormattingRule(
                            style={
                                "color": WBColor.GREEN_DARK.value,
                                "fontWeight": "bold",
                            },
                            condition=(">=", 0),
                        ),
                    ],
                )
                if self.view.dates
                else {},
            ]
        )



from datetime import date

from wbportfolio.pms.typing import Portfolio as PortfolioDTO


class AbstractRebalancingModel:
    @property
    def validation_errors(self) -> str:
        return getattr(self, "_validation_errors", "Rebalacing cannot applied for these parameters")

    def __init__(self, portfolio, trade_date: date, last_effective_date: date, **kwargs):
        self.portfolio = portfolio
        self.trade_date = trade_date
        self.last_effective_date = last_effective_date

    def is_valid(self) -> bool:
        return True

    def get_target_portfolio(self) -> PortfolioDTO:
        raise NotImplementedError()



# Generated by Django 4.1.9 on 2023-07-19 07:11

from decimal import Decimal

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbportfolio", "0042_instrumentlist_instrumentlistthroughmodel_and_more"),
        ("currency", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="instrumentlistthroughmodel",
            options={"verbose_name": "Instrument in Instrument List"},
        ),
        migrations.RemoveField(
            model_name="instrumentlist",
            name="instruments",
        ),
        migrations.AddField(
            model_name="instrumentlistthroughmodel",
            name="instrument_str",
            field=models.CharField(default="", max_length=256),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="instrumentlistthroughmodel",
            name="instrument",
            field=models.ForeignKey(
                blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to="wbportfolio.instrument"
            ),
        ),
        migrations.AlterUniqueTogether(
            name="instrumentprice",
            unique_together={("instrument", "date", "calculated")},
        ),
        migrations.AddField(
            model_name="instrumentprice",
            name="currency_fx_rate_to_usd",
            field=models.ForeignKey(
                blank=True,
                help_text="Rate to between instrument currency and USD",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="instrument_prices",
                to="currency.currencyfxrates",
                verbose_name="Instrument Currency Rate",
            ),
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="ask",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="bid",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="bid_rate",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="close",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="currency_fx_usd",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="custom_beta_180d",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="custom_beta_1y",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="custom_beta_2y",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="custom_beta_3y",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="custom_beta_5y",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="dividend_yield",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="free_float",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="free_float_ratio",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="high",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="interest_rate",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="low",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="market_capitalization_usd",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="net_return",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="net_value_usd",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="offered_rate",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="open",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="performance_1d",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="performance_30d",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="performance_365d",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="performance_7d",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="performance_90d",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="performance_inception",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="performance_ytd",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="price_earning",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="price_index",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="short_interest",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="short_interest_ratio",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="total_return_index",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="volume_50d_usd",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="volume_usd",
        ),
        migrations.RemoveField(
            model_name="instrumentprice",
            name="yield_redemption",
        ),
        migrations.RenameModel(old_name="SplitData", new_name="Adjustment"),
        migrations.RenameField(model_name="assetposition", old_name="applied_split", new_name="applied_adjustment"),
        migrations.AlterModelOptions(
            name="adjustment",
            options={"verbose_name": "Datum Adjustment", "verbose_name_plural": "Data Adjustment"},
        ),
        migrations.AlterField(
            model_name="adjustment",
            name="cumulative_factor",
            field=models.DecimalField(
                decimal_places=17, default=Decimal("1"), max_digits=30, verbose_name="Cumulative Factor"
            ),
        ),
        migrations.AlterField(
            model_name="adjustment",
            name="factor",
            field=models.DecimalField(decimal_places=17, default=Decimal("1"), max_digits=30, verbose_name="Factor"),
        ),
        migrations.AlterField(
            model_name="adjustment",
            name="last_handler",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="last_handled_adjustments",
                to="directory.person",
            ),
        ),
        migrations.AlterField(
            model_name="assetposition",
            name="applied_adjustment",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="adjustmented_assets",
                to="wbportfolio.adjustment",
            ),
        ),
    ]



# Generated by Django 4.2.8 on 2023-12-18 13:46

import django.db.models.deletion
from django.db import migrations, models


def migrate_productgroup_to_parent(apps, schema_editor):
    ProductGroup = apps.get_model("wbportfolio", "ProductGroup")
    Product = apps.get_model("wbportfolio", "Product")
    Instrument = apps.get_model("wbfdm", "Instrument")
    for product_group in ProductGroup.objects.all():
        for product in Product.objects.filter(group=product_group):
            instrument = Instrument.objects.get(id=product.instrument_ptr_id)
            instrument.parent_id = product_group.instrument_ptr_id
            instrument.save()


class Migration(migrations.Migration):
    dependencies = [
        ("wbfdm", "0012_instrumentprice_created_instrumentprice_modified"),
        ("wbportfolio", "0052_remove_cash_instrument_ptr_and_more"),
    ]

    operations = [
        migrations.RunPython(migrate_productgroup_to_parent),
        migrations.RemoveField(
            model_name="product",
            name="group",
        ),
        migrations.AlterField(
            model_name="adjustment",
            name="instrument",
            field=models.ForeignKey(
                limit_choices_to=models.Q(("children__isnull", True)),
                on_delete=django.db.models.deletion.CASCADE,
                related_name="pms_adjustments",
                to="wbfdm.instrument",
            ),
        ),
        migrations.AlterField(
            model_name="assetposition",
            name="underlying_instrument",
            field=models.ForeignKey(
                help_text="The instrument that is this asset.",
                limit_choices_to=models.Q(("children__isnull", True)),
                on_delete=django.db.models.deletion.PROTECT,
                related_name="assets",
                to="wbfdm.instrument",
                verbose_name="Underlying Instrument",
            ),
        ),
        migrations.AlterField(
            model_name="portfoliorole",
            name="instrument",
            field=models.ForeignKey(
                blank=True,
                limit_choices_to=models.Q(("children__isnull", True)),
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="portfolio_roles",
                to="wbfdm.instrument",
                verbose_name="Instrument",
            ),
        ),
        migrations.AlterField(
            model_name="transaction",
            name="underlying_instrument",
            field=models.ForeignKey(
                help_text="The instrument that is this transaction.",
                limit_choices_to=models.Q(("children__isnull", True)),
                on_delete=django.db.models.deletion.PROTECT,
                related_name="transactions",
                to="wbfdm.instrument",
                verbose_name="Underlying Instrument",
            ),
        ),
        migrations.AlterField(
            model_name="portfolio",
            name="prefered_instrument_classifications",
            field=models.ManyToManyField(
                blank=True,
                limit_choices_to=models.Q(("instrument_type__is_classifiable", True)),
                related_name="prefered_portfolio_classifications",
                through="wbportfolio.PortfolioInstrumentPreferedClassificationThroughModel",
                to="wbfdm.instrument",
                verbose_name="The Prefered classification per instrument",
            ),
        ),
        migrations.AlterField(
            model_name="portfolioinstrumentpreferedclassificationthroughmodel",
            name="instrument",
            field=models.ForeignKey(
                limit_choices_to=models.Q(("instrument_type__is_classifiable", True)),
                on_delete=django.db.models.deletion.CASCADE,
                related_name="prefered_classification_portfolio",
                to="wbfdm.instrument",
            ),
        ),
        migrations.AlterField(
            model_name="instrumentportfoliothroughmodel",
            name="instrument",
            field=models.ForeignKey(
                limit_choices_to=models.Q(
                    ("instrument_type__key", "product"),
                    ("instrument_type__key", "index"),
                    ("instrument_type__key", "product_group"),
                    _connector="OR",
                ),
                on_delete=django.db.models.deletion.CASCADE,
                related_name="through_portfolios",
                to="wbfdm.instrument",
            ),
        ),
        migrations.AlterField(
            model_name="portfolio",
            name="instruments",
            field=models.ManyToManyField(
                blank=True,
                help_text="Instruments linked to this instrument",
                limit_choices_to=models.Q(
                    ("instrument_type__key", "product"),
                    ("instrument_type__key", "index"),
                    ("instrument_type__key", "product_group"),
                    _connector="OR",
                ),
                related_name="portfolios",
                through="wbportfolio.InstrumentPortfolioThroughModel",
                to="wbfdm.instrument",
                verbose_name="Instruments",
            ),
        ),
    ]



from decimal import Decimal

from django.apps import apps
from django.db import models
from django.dispatch import receiver
from wbcore.contrib.io.mixins import ImportMixin
from wbcore.signals import pre_merge
from wbfdm.models.instruments.instruments import Instrument
from wbfdm.signals import add_instrument_to_investable_universe


class ShareMixin(models.Model):
    shares = models.DecimalField(
        max_digits=15,
        decimal_places=4,
        default=Decimal("0.0"),
        help_text="The number of shares that were traded.",
        verbose_name="Shares",
    )
    price = models.DecimalField(
        max_digits=16,
        decimal_places=4,
        default=Decimal(
            "0.0"
        ),  # we shouldn't default to anything but we have trade with price=None. Needs to be handled carefully
        help_text="The price per share.",
        verbose_name="Price",
    )

    price_gross = models.DecimalField(
        max_digits=16,
        decimal_places=4,
        help_text="The gross price per share.",
        verbose_name="Gross Price",
    )

    def save(
        self,
        *args,
        factor: Decimal = Decimal("1.0"),
        **kwargs,
    ):
        if self.price_gross is None:
            self.price_gross = self.price

        self.total_value = self.price * self.shares * factor
        self.total_value_gross = self.price_gross * self.shares * factor
        super().save(*args, **kwargs)

    class Meta:
        abstract = True


class Transaction(ImportMixin, models.Model):
    class Type(models.TextChoices):
        # Standart Asset Types
        TRADE = "Trade", "Trade"
        DIVIDEND_TRANSACTION = "DividendTransaction", "Dividend Transaction"
        EXPIRY = "Expiry", "Expiry"
        FEES = "Fees", "Fees"

    transaction_type = models.CharField(max_length=255, verbose_name="Type", choices=Type.choices, default=Type.TRADE)

    portfolio = models.ForeignKey(
        "wbportfolio.Portfolio", related_name="transactions", on_delete=models.PROTECT, verbose_name="Portfolio"
    )

    underlying_instrument = models.ForeignKey(
        to="wbfdm.Instrument",
        related_name="transactions",
        limit_choices_to=models.Q(children__isnull=True),
        on_delete=models.PROTECT,
        verbose_name="Underlying Instrument",
        help_text="The instrument that is this transaction.",
    )

    transaction_date = models.DateField(
        verbose_name="Trade Date",
        help_text="The date that this transaction was traded.",
    )
    book_date = models.DateField(
        verbose_name="Trade Date",
        help_text="The date that this transaction was booked.",
    )
    value_date = models.DateField(
        verbose_name="Value Date",
        help_text="The date that this transaction was valuated.",
    )

    currency = models.ForeignKey(
        "currency.Currency",
        related_name="transactions",
        on_delete=models.PROTECT,
        verbose_name="Currency",
    )
    currency_fx_rate = models.DecimalField(
        max_digits=14, decimal_places=8, default=Decimal(1.0), verbose_name="FOREX rate"
    )
    total_value = models.DecimalField(max_digits=20, decimal_places=4, verbose_name="Total Value")
    total_value_gross = models.DecimalField(max_digits=20, decimal_places=4, verbose_name="Total Value Gross")
    total_value_fx_portfolio = models.GeneratedField(
        expression=models.F("currency_fx_rate") * models.F("total_value"),
        output_field=models.DecimalField(
            max_digits=20,
            decimal_places=4,
        ),
        db_persist=True,
    )
    total_value_gross_fx_portfolio = models.GeneratedField(
        expression=models.F("currency_fx_rate") * models.F("total_value_gross"),
        output_field=models.DecimalField(
            max_digits=20,
            decimal_places=4,
        ),
        db_persist=True,
    )
    external_id = models.CharField(
        max_length=255,
        null=True,
        blank=True,
        help_text="An external identifier that was supplied.",
        verbose_name="External Identifier",
    )
    comment = models.TextField(default="", verbose_name="Comment", blank=True)

    def save(self, *args, **kwargs):
        if not self.value_date:
            self.value_date = self.transaction_date
        if not self.book_date:
            self.book_date = self.transaction_date

        if not getattr(self, "currency", None) and self.underlying_instrument:
            self.currency = self.underlying_instrument.currency
        if self.currency_fx_rate is None:
            self.currency_fx_rate = self.underlying_instrument.currency.convert(
                self.value_date, self.portfolio.currency, exact_lookup=True
            )
        if not self.transaction_type:
            self.transaction_type = self.__class__.__name__

        if self.total_value_gross is None:
            self.total_value_gross = self.total_value

        super().save(*args, **kwargs)

    def __str__(self):
        return f"{self.total_value} - {self.transaction_date:%d.%m.%Y} : {str(self.underlying_instrument)} (in {str(self.portfolio)})"

    def get_casted_model(self):
        return apps.get_model(app_label="wbportfolio", model_name=self.transaction_type)

    def get_casted_transaction(self) -> models.Model:
        """
        Cast the asset into its child representative
        """
        model = self.get_casted_model()
        return model.objects.get(pk=self.pk)

    class Meta:
        verbose_name = "Transaction"
        verbose_name_plural = "Transactions"
        indexes = [
            models.Index(fields=["underlying_instrument", "transaction_date"]),
            # models.Index(fields=["date", "underlying_instrument"]),
        ]

    objects = models.Manager()

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{total_value}}{{transaction_date}}"

    @classmethod
    def get_endpoint_basename(cls):
        return "wbportfolio:transaction"


@receiver(pre_merge, sender="wbfdm.Instrument")
def pre_merge_instrument(sender: models.Model, merged_object: "Instrument", main_object: "Instrument", **kwargs):
    """
    Simply reassign the transactions linked to the merged instrument to the main instrument
    """
    merged_object.transactions.update(underlying_instrument=main_object)


@receiver(add_instrument_to_investable_universe, sender="wbfdm.Instrument")
def add_instrument_to_investable_universe_from_transactions(sender: models.Model, **kwargs) -> list[int]:
    """
    register all instrument linked to assets as within the investible universe
    """
    return list(
        (
            Instrument.objects.annotate(
                transaction_exists=models.Exists(
                    Transaction.objects.filter(underlying_instrument=models.OuterRef("pk"))
                )
            ).filter(transaction_exists=True)
        )
        .distinct()
        .values_list("id", flat=True)
    )



import xlrd

from wbportfolio.models import Product, Trade


def get_isin(sheet):
    for row in sheet.get_rows():
        if row[0].value == "ISIN":
            return row[1].value


def parse(import_source):
    book = xlrd.open_workbook(file_contents=import_source.file.read())

    product = Product.objects.get(isin=get_isin(book.sheet_by_name("Certificate")))
    valuation_sheet = book.sheet_by_name("Transaction Fees")

    trade_date_column = None
    shares_column = None
    price_column = None
    currency_fx_rate_column = None
    isin_column = None
    execution_fees_column = None

    data = list()
    for i, row in enumerate(valuation_sheet.get_rows()):
        if i == 0:
            for column in range(0, valuation_sheet.ncols):
                if "Trade date" == row[column].value:
                    trade_date_column = column

                if "Quantity" == row[column].value:
                    shares_column = column

                if "Price" == row[column].value:
                    price_column = column

                if "Forex" == row[column].value:
                    currency_fx_rate_column = column

                if "Reference" == row[column].value:
                    isin_column = column

                if "Exec Fees" in row[column].value:
                    execution_fees_column = column

        else:
            transaction_date = xlrd.xldate.xldate_as_datetime(row[trade_date_column].value, 0)
            shares = row[shares_column].value
            price = row[price_column].value
            currency_fx_rate = row[currency_fx_rate_column].value
            row[execution_fees_column].value

            traded_product = Product.objects.get(isin=row[isin_column].value)

            data.append(
                {
                    "portfolio": product.primary_portfolio.id,
                    "transaction_subtype": Trade.Type.REBALANCE,
                    "transaction_date": transaction_date.strftime("%Y-%m-%d"),
                    "shares": shares,
                    "price": price,
                    "currency__key": traded_product.currency.key,
                    "currency_fx_rate": currency_fx_rate,
                    "underlying_instrument": {"id": traded_product.id, "instrument_type": "product"},
                    # 'execution_fees_percent': 0,
                    # 'market_fees_percent': 0,
                }
            )

    return {"data": data}



from datetime import datetime

import pandas as pd

from wbportfolio.models import ProductGroup


def parse(import_source):
    csv_file = import_source.file.open()

    df = pd.read_csv(csv_file, encoding="latin1", usecols=[0, 2, 4, 18])

    value_date = df["NAV Date"].iloc[0]
    total_assets = df.groupby("Code").sum("Market Value FCY")
    cash = df[df["Asset type code"] == "TRES"].groupby("Code").sum("Market Value FCY")

    data = []

    for code in df["Code"].unique():
        data.append(
            {
                "portfolio": ProductGroup.objects.get(identifier=code).portfolio.id,
                "value_date": datetime.strptime(value_date, "%Y/%m/%d").strftime("%Y-%m-%d"),
                "total_assets": total_assets.T[code].iloc[0],
                "cash": cash.T[code].iloc[0],
                "pending": False,
            }
        )

    return {"data": data}



from wbcore.metadata.configs.titles import TitleViewConfig

from wbportfolio.models import Portfolio


class FeesTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Fees"

    def get_instance_title(self):
        return "Fees: {{_product.name}} {{date}}"

    def get_create_title(self):
        return "New Fees"


class FeesPortfolioTitleConfig(FeesTitleConfig):
    def get_list_title(self):
        portfolio = Portfolio.objects.get(id=self.view.kwargs["portfolio_id"])
        return f"Fees: {portfolio.name}"


class FeesAggregatedPortfolioTitleConfig(TitleViewConfig):
    def get_list_title(self):
        portfolio = Portfolio.objects.get(id=self.view.kwargs["portfolio_id"])
        return f"Aggregated Fees for {portfolio.name}"



import datetime
import re

import numpy as np
import pandas as pd

from wbportfolio.import_export.utils import convert_string_to_number
from wbportfolio.models import Product, Trade


def file_name_parse(file_name):
    dates = re.findall("([0-9]{8})", file_name)

    assert len(dates) > 0

    parts_dict = {"valuation_date": datetime.datetime.strptime(dates[0], "%Y%m%d").date()}

    isin = re.findall("([A-Z]{2}[A-Z0-9]{9}[0-9]{1})", file_name)

    if len(isin) > 0:
        parts_dict["isin"] = isin[0]

    return parts_dict


def parse(import_source):
    bank_exclusion_list = [
        excluded_bank.strip().lower()
        for excluded_bank in import_source.source.import_parameters.get("bank_exclusion_list", [])
    ]
    # Load files into a CSV DictReader
    df = pd.read_csv(import_source.file, encoding="latin1", delimiter=",")
    df = df.replace([np.inf, -np.inf, np.nan], None)
    # Parse the Parts of the filename into the different parts
    parts = file_name_parse(import_source.file.name)
    df["Trade Date"] = pd.to_datetime(df["Trade Date"], format="%Y%m%d")
    # Get the valuation date from the parts list
    valuation_date = parts["valuation_date"]

    # Iterate through the CSV File and parse the data into a list
    data = list()
    for nominal_data in df.to_dict("records"):
        product = Product.objects.get(isin=nominal_data["ISIN"])
        shares = convert_string_to_number(nominal_data["Quantity"])

        # Check whether it is a buy or a sell and convert the value correspondely
        shares = shares if nominal_data["Side"] == "S" else shares * -1
        portfolio = product.primary_portfolio
        bank = nominal_data["CounterParty Name"]
        if (
            len(bank_exclusion_list) == 0 or bank.strip().lower() not in bank_exclusion_list
        ):  # we do basic string comparison to exclude appropriate banks. We might want to include regex if bank data is inconsistent
            data.append(
                {
                    "underlying_instrument": {"id": product.id, "instrument_type": "product"},
                    "transaction_date": nominal_data["Trade Date"].strftime("%Y-%m-%d"),
                    "shares": shares,
                    "portfolio": portfolio.id,
                    # 'currency': product.currency.key,
                    "transaction_subtype": Trade.Type.REDEMPTION if shares < 0 else Trade.Type.SUBSCRIPTION,
                    "bank": bank,
                    "price": convert_string_to_number(nominal_data["Price"]),
                }
            )
    import_data = {"data": data}
    if "isin" in parts:
        product = Product.objects.get(isin=parts["isin"])
        import_data["history"] = {
            "underlying_instrument": product.id,
            "transaction_date": valuation_date.strftime("%Y-%m-%d"),
        }
    return import_data



import xlrd

from wbportfolio.models import Product


def get_isin(sheet):
    for row in sheet.get_rows():
        if row[0].value == "ISIN":
            return row[1].value


def parse(import_source):
    book = xlrd.open_workbook(file_contents=import_source.file.read())

    product = Product.objects.get(isin=get_isin(book.sheet_by_name("Certificate")))
    valuation_sheet = book.sheet_by_name("Report")

    date_column = None
    net_value_column = None

    data = list()
    for i, row in enumerate(valuation_sheet.get_rows()):
        if i == 0:
            for column in range(0, valuation_sheet.ncols):
                if "Date" in row[column].value:
                    date_column = column
                if "price" in row[column].value:
                    net_value_column = column
        else:
            valuation_date = xlrd.xldate.xldate_as_datetime(row[date_column].value, 0)
            net_value = row[net_value_column].value

            data.append(
                {
                    "instrument": {"instrument_type": "product", "id": product.id},
                    "date": valuation_date.strftime("%Y-%m-%d"),
                    "net_value": net_value,
                    "calculated": False,
                }
            )

    return {"data": data}



from datetime import datetime, timedelta

from wbcore.contrib.io.exceptions import DeserializationError
from wbcore.contrib.io.imports import ImportExportHandler
from wbfdm.import_export.handlers.instrument import InstrumentImportHandler


class AdjustmentImportHandler(ImportExportHandler):
    MODEL_APP_LABEL = "wbportfolio.Adjustment"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.instrument_handler = InstrumentImportHandler(self.import_source)

    def _deserialize(self, data):
        data["date"] = datetime.strptime(data["date"], "%Y-%m-%d").date()
        data["instrument"] = self.instrument_handler.process_object(
            data["instrument"], only_security=True, read_only=True
        )[0]
        if data["factor"] is None:
            raise DeserializationError("Can't process this data: no factor")

    def _get_instance(self, data, history=None, **kwargs):
        self.import_source.log += "\nGet Adjustment Instance."
        self.import_source.log += f"\nParameter: Instrument={data['instrument']} Date={data['date']}"

        if exact_adjustment := data["instrument"].pms_adjustments.filter(date=data["date"]).first():
            return exact_adjustment
        potential_adjustments = data["instrument"].pms_adjustments.filter(
            date__gte=data["date"] - timedelta(days=7),
            date__lte=data["date"] + timedelta(days=7),
            factor=data["factor"],
        )
        if potential_adjustments.exists():
            return potential_adjustments.first()

    def _create_instance(self, data, **kwargs):
        self.import_source.log += "\nCreate Adjustment datapoint."
        return self.model.objects.create(**data, import_source=self.import_source)



from django.db.models import Case, F, Value, When
from wbcore import viewsets

from wbportfolio.models import DailyPortfolioCashFlow
from wbportfolio.serializers import DailyPortfolioCashFlowModelSerializer
from wbportfolio.viewsets.configs.display import DailyPortfolioCashFlowDisplayConfig
from wbportfolio.viewsets.configs.titles import DailyPortfolioCashFlowTitleConfig


class DailyPortfolioCashFlowModelViewSet(viewsets.ModelViewSet):
    queryset = DailyPortfolioCashFlow.objects.all()
    serializer_class = DailyPortfolioCashFlowModelSerializer

    display_config_class = DailyPortfolioCashFlowDisplayConfig
    title_config_class = DailyPortfolioCashFlowTitleConfig

    ordering_fields = ("value_date",)
    ordering = ("-value_date",)

    def get_queryset(self):
        queryset = super().get_queryset()

        if portfolio_id := self.kwargs.get("portfolio_id", None):
            queryset = queryset.filter(portfolio_id=portfolio_id)

        return queryset.annotate(
            swing_pricing_indicator=Case(
                When(cash_flow_asset_ratio__lt=F("swing_pricing__negative_threshold"), then=Value("neg")),
                When(cash_flow_asset_ratio__gt=F("swing_pricing__positive_threshold"), then=Value("pos")),
                default=None,
            )
        )



from contextlib import suppress

from django.db import models
from django.db.models.signals import post_save
from django.dispatch import receiver

# from django.utils.translation import gettext_lazy as _
from dynamic_preferences.registries import global_preferences_registry
from wbcore.contrib.authentication.models import User
from wbcore.contrib.notifications.dispatch import send_notification
from wbcore.contrib.notifications.utils import create_notification_type

from wbportfolio.models.reconciliations.account_reconciliation_lines import (
    AccountReconciliationLine,
)
from wbportfolio.models.reconciliations.reconciliations import Reconciliation


class AccountReconciliationQuerySet(models.QuerySet):
    def create(self, **kwargs) -> models.QuerySet["AccountReconciliation"]:
        reconciliation = super().create(**kwargs)
        AccountReconciliationLine.objects.update_or_create_for_reconciliation(reconciliation)
        return reconciliation

    def filter_for_user(self, user: User) -> models.QuerySet["AccountReconciliation"]:
        if user.is_superuser or user.is_internal:
            return self

        if not user.has_perm("wbportfolio.view_accountreconciliation"):
            return self.none()

        return self.filter(
            account__roles__role_type__key="reconciliation-manager",
            account__roles__entry_id=user.profile_id,
        )


class AccountReconciliation(Reconciliation):
    lines: models.QuerySet[AccountReconciliationLine]

    account = models.ForeignKey(
        to="wbcrm.Account",
        related_name="wbportfolio_reconciliations",
        on_delete=models.CASCADE,
    )

    objects = AccountReconciliationQuerySet.as_manager()

    class Meta:
        verbose_name = "Account Holdings Reconciliation"
        verbose_name_plural = "Account Holdings Reconciliations"
        constraints = [
            models.UniqueConstraint(fields=["reconciliation_date", "account"], name="unique_date_account"),
        ]
        notification_types = [
            create_notification_type(
                "wbportfolio.account_reconciliation.notify",
                "Account Reconciliation Events",
                "A notification that informs the user about an account reconciliation event.",
                email=True,
            ),
        ]

    def __str__(self) -> str:
        return f"{self.account} ({self.reconciliation_date:%d.%m.%Y})"

    @classmethod
    def get_endpoint_basename(cls):
        return "wbportfolio:accountreconciliation"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbportfolio:accountreconciliationrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{account}} {{reconciliation_date}}"

    def notify(self, update: bool = False):
        registry = global_preferences_registry.manager()
        title = registry["wbportfolio__account_holding_reconciliation_notification_title"]
        body = registry[f"wbportfolio__account_holding_reconciliation_notification_body{'_update' if update else ''}"]
        for role in self.account.roles.filter(role_type__key="reconciliation-manager"):
            with suppress(User.DoesNotExist):
                user = User.objects.get(profile_id=role.entry.id)
                send_notification(
                    code="wbportfolio.account_reconciliation.notify",
                    title=title,
                    body=body.format(account=self.account, reconciliation_date=self.reconciliation_date),
                    user=user,
                    reverse_name="wbportfolio:accountreconciliation-detail",
                    reverse_args=[self.id],
                )


@receiver(post_save, sender=AccountReconciliation)
def notify_on_create(sender, instance: AccountReconciliation, created: bool, **kwargs):
    if created:
        instance.notify()



from wbfdm.models import Cash, InstrumentPrice

from wbportfolio.models import FeeProductPercentage, Fees, Product


def fees_calculation(price_id):
    price = InstrumentPrice.objects.get(id=price_id)
    if price.calculated:
        raise ValueError("Cannot compute fees on calculated price")
    currency = price.instrument.currency
    underlying_instrument = Cash.objects.filter(currency=currency).first()
    product = Product.objects.get(id=price.instrument.id)
    portfolio = product.portfolio
    previous_price = price.previous_price

    if previous_price:
        shares = product.total_shares(previous_price.date)
        multiplicator = shares * previous_price.net_value
    else:
        shares = product.total_shares(price.date)
        multiplicator = shares * product.share_price

    product_net_management_fees = product.get_fees_percent(price.date, FeeProductPercentage.Type.MANAGEMENT)
    product_net_performance_fees = product.get_fees_percent(price.date, FeeProductPercentage.Type.PERFORMANCE)
    product_gross_performance_fees = product.get_fees_percent(
        price.date, FeeProductPercentage.Type.PERFORMANCE, net=False
    )

    if previous_price and previous_price.date.weekday() == 4:
        # The previous InstrumentPrice was a Friday. This means we have to calculate for the weekend
        previous_total_value = shares * previous_price.net_value

        sat_bank_fees = (product.bank_fees / 360) * previous_total_value
        sat_management_fees = (product_net_management_fees / 360) * previous_total_value
        previous_total_value = previous_total_value - sat_bank_fees - sat_management_fees

        sun_bank_fees = (product.bank_fees / 360) * previous_total_value
        sun_management_fees = (product_net_management_fees / 360) * previous_total_value
        previous_total_value = previous_total_value - sun_bank_fees - sun_management_fees

        mon_bank_fees = (product.bank_fees / 360) * previous_total_value
        mon_management_fees = (product_net_management_fees / 360) * previous_total_value
        bank_fees = sat_bank_fees + sun_bank_fees + mon_bank_fees
        management_fees = sat_management_fees + sun_management_fees + mon_management_fees
    else:
        bank_fees = (product.bank_fees / 360) * multiplicator
        management_fees = (product_net_management_fees / 360) * multiplicator

    value = (
        price.net_value or price.gross_value
    )  # Flipped around, so that if we have a good net price from the bank, we rather use that one. Issue is, that if the gross price is completely off, the performance fees can be really really high.
    multiplicator = max(0, value - product.get_high_water_mark(price.date)) * shares
    performance_fees_net = product_net_performance_fees * multiplicator
    performance_fees_gross = product_gross_performance_fees * multiplicator
    base_fields = [
        "total_value",
        "total_value_fx_portfolio",
        "total_value_gross",
        "total_value_gross_fx_portfolio",
    ]
    yield {
        "portfolio": portfolio,
        "linked_product": product,
        "transaction_date": price.date,
        "transaction_subtype": Fees.Type.MANAGEMENT,
        "underlying_instrument": underlying_instrument,
        "currency": currency,
        "calculated": True,
        **{f: management_fees for f in base_fields},
    }
    yield {
        "portfolio": portfolio,
        "linked_product": product,
        "transaction_date": price.date,
        "transaction_subtype": Fees.Type.ISSUER,
        "underlying_instrument": underlying_instrument,
        "currency": currency,
        "calculated": True,
        **{f: bank_fees for f in base_fields},
    }
    yield {
        "portfolio": portfolio,
        "linked_product": product,
        "transaction_date": price.date,
        "transaction_subtype": Fees.Type.PERFORMANCE,
        "underlying_instrument": underlying_instrument,
        "currency": currency,
        "calculated": True,
        "total_value": performance_fees_net,
        "total_value_fx_portfolio": performance_fees_net,
        "total_value_gross": performance_fees_gross,
        "total_value_gross_fx_portfolio": performance_fees_gross,
    }



from datetime import date
from decimal import Decimal
from typing import TYPE_CHECKING

import pytest
from django.db.utils import IntegrityError

if TYPE_CHECKING:
    from wbportfolio.factories import DailyPortfolioCashFlowFactory
    from wbportfolio.models.portfolio import Portfolio
    from wbportfolio.models.portfolio_cash_flow import DailyPortfolioCashFlow
    from wbportfolio.models.portfolio_cash_targets import PortfolioCashTarget


@pytest.mark.django_db
class TestDailyPortfolioCashFlowFactory:
    def test_factory(self, daily_portfolio_cash_flow: "DailyPortfolioCashFlow"):
        assert daily_portfolio_cash_flow.pk is not None

    def test_constraint_unique_value_date_portfolio(
        self, portfolio: "Portfolio", daily_portfolio_cash_flow_factory: "DailyPortfolioCashFlowFactory"
    ):
        cf1 = daily_portfolio_cash_flow_factory.create(portfolio=portfolio)
        cf2 = daily_portfolio_cash_flow_factory.create(portfolio=portfolio)

        cf2.value_date = cf1.value_date
        with pytest.raises(IntegrityError):
            cf2.save()

    @pytest.mark.parametrize("daily_portfolio_cash_flow__pending", [False])
    def test_estimated_total_assets(self, daily_portfolio_cash_flow: "DailyPortfolioCashFlow"):
        dpcf = daily_portfolio_cash_flow
        assert dpcf.estimated_total_assets == dpcf.total_assets

    @pytest.mark.parametrize("daily_portfolio_cash_flow__pending", [True])
    def test_estimated_total_assets_while_pending(self, daily_portfolio_cash_flow: "DailyPortfolioCashFlow"):
        dpcf = daily_portfolio_cash_flow
        assert dpcf.estimated_total_assets == dpcf.total_assets + dpcf.cash_flow_forecast

    def test_cash_flow_asset_ratio(self, daily_portfolio_cash_flow: "DailyPortfolioCashFlow"):
        dpcf = daily_portfolio_cash_flow
        assert dpcf.cash_flow_asset_ratio == pytest.approx(
            dpcf.cash_flow_forecast / dpcf.estimated_total_assets, rel=Decimal(1e-4)
        )

    @pytest.mark.parametrize("daily_portfolio_cash_flow__total_assets", [Decimal(0)])
    @pytest.mark.parametrize("daily_portfolio_cash_flow__cash_flow_forecast", [Decimal(0)])
    def test_cash_flow_asset_ratio_zero_assets(self, daily_portfolio_cash_flow: "DailyPortfolioCashFlow"):
        assert daily_portfolio_cash_flow.cash_flow_asset_ratio == Decimal(0)

    def test_cash_pct(self, daily_portfolio_cash_flow: "DailyPortfolioCashFlow"):
        dpct = daily_portfolio_cash_flow
        assert dpct.cash_pct == pytest.approx(dpct.cash / dpct.estimated_total_assets, rel=Decimal(1e-4))

    @pytest.mark.parametrize("daily_portfolio_cash_flow__total_assets", [Decimal(0)])
    @pytest.mark.parametrize("daily_portfolio_cash_flow__cash_flow_forecast", [Decimal(0)])
    def test_cash_pct_zero_assets(self, daily_portfolio_cash_flow: "DailyPortfolioCashFlow"):
        assert daily_portfolio_cash_flow.cash_pct == Decimal(0)

    def test_true_cash(self, daily_portfolio_cash_flow: "DailyPortfolioCashFlow"):
        dpct = daily_portfolio_cash_flow
        assert dpct.true_cash == pytest.approx(dpct.cash + dpct.cash_flow_forecast, rel=Decimal(1e-4))

    def test_true_cash_pct(self, daily_portfolio_cash_flow: "DailyPortfolioCashFlow"):
        dpct = daily_portfolio_cash_flow
        assert dpct.true_cash_pct == pytest.approx(dpct.true_cash / dpct.estimated_total_assets, rel=Decimal(1e-4))

    @pytest.mark.parametrize("daily_portfolio_cash_flow__total_assets", [Decimal(0)])
    @pytest.mark.parametrize("daily_portfolio_cash_flow__cash_flow_forecast", [Decimal(0)])
    def test_true_cash_pct_zero_assets(self, daily_portfolio_cash_flow: "DailyPortfolioCashFlow"):
        assert daily_portfolio_cash_flow.true_cash_pct == Decimal(0)

    def test_target_cash_no_target(self, daily_portfolio_cash_flow: "DailyPortfolioCashFlow"):
        dpct = daily_portfolio_cash_flow
        assert dpct.target_cash == Decimal(0)

    def test_target_cash(
        self, daily_portfolio_cash_flow: "DailyPortfolioCashFlow", portfolio_cash_target: "PortfolioCashTarget"
    ):
        dpct = daily_portfolio_cash_flow
        portfolio_cash_target.portfolio = dpct.portfolio
        portfolio_cash_target.valid_date = dpct.value_date
        portfolio_cash_target.save()
        dpct.save()
        dpct.refresh_from_db()
        assert dpct.target_cash == pytest.approx(portfolio_cash_target.target * dpct.estimated_total_assets)

    def test_excess_cash(
        self, daily_portfolio_cash_flow: "DailyPortfolioCashFlow", portfolio_cash_target: "PortfolioCashTarget"
    ):
        dpct = daily_portfolio_cash_flow
        portfolio_cash_target.portfolio = dpct.portfolio
        portfolio_cash_target.valid_date = dpct.value_date
        portfolio_cash_target.save()
        dpct.save()
        dpct.refresh_from_db()
        assert dpct.excess_cash == pytest.approx(dpct.true_cash - dpct.target_cash, rel=Decimal(1e-4))

    def test_cash_from_yesterday(
        self, portfolio: "Portfolio", daily_portfolio_cash_flow_factory: "DailyPortfolioCashFlowFactory"
    ):
        cf1 = daily_portfolio_cash_flow_factory.create(portfolio=portfolio, value_date=date(2020, 1, 1))
        cf2 = daily_portfolio_cash_flow_factory.create(
            portfolio=portfolio, value_date=date(2020, 1, 2), pending=True, rebalancing=Decimal(0)
        )
        cf1.refresh_from_db()
        cf2.refresh_from_db()
        assert cf2.cash == pytest.approx(cf1.true_cash, rel=Decimal(1e-4))

    @pytest.mark.parametrize("daily_portfolio_cash_flow__cash", [Decimal(1_000_000)])
    @pytest.mark.parametrize("daily_portfolio_cash_flow__pending", [True])
    def test_rebalancing(self, daily_portfolio_cash_flow: "DailyPortfolioCashFlow"):
        dpcf = daily_portfolio_cash_flow
        assert dpcf.cash == Decimal(1_000_000) - dpcf.rebalancing



import datetime

import xlrd

from wbportfolio.models import Product, Trade


def get_isin(sheet):
    for row in sheet.get_rows():
        if row[0].value == "ISIN":
            return row[1].value


def parse(import_source):
    book = xlrd.open_workbook(file_contents=import_source.file.read())

    product_sheet = book.sheet_by_name("Certificate")
    valuation_sheet = book.sheet_by_name("Report")

    product = Product.objects.get(isin=get_isin(product_sheet))

    date_column = None
    net_value_column = None
    nominal_column = None

    max_date = datetime.date(1900, 1, 1)

    data = list()
    for i, row in enumerate(valuation_sheet.get_rows()):
        if i == 0:
            for column in range(0, valuation_sheet.ncols):
                if "Date" in row[column].value:
                    date_column = column
                if "price" in row[column].value:
                    net_value_column = column
                if "Nominal" == row[column].value:
                    nominal_column = column

        else:
            transaction_date = xlrd.xldate.xldate_as_datetime(row[date_column].value, 0).date()
            max_date = max(max_date, transaction_date)
            net_value = row[net_value_column].value

            shares_today = row[nominal_column].value / product.share_price

            try:
                shares_yesterday = valuation_sheet.cell_value(i + 1, nominal_column) / product.share_price
                trade = shares_today - shares_yesterday
            except IndexError:
                trade = shares_today

            if trade and trade != 0:
                data.append(
                    {
                        "underlying_instrument": {"id": product.id, "instrument_type": "product"},
                        "transaction_date": transaction_date.strftime("%Y-%m-%d"),
                        "shares": trade,
                        "transaction_subtype": Trade.Type.REDEMPTION if trade < 0 else Trade.Type.SUBSCRIPTION,
                        "bank": "Natixis Internal Trade",
                        "price": net_value,
                    }
                )

    return {
        "data": data,
        "history": {"underlying_instrument": product.id, "transaction_date": max_date.strftime("%Y-%m-%d")},
    }



from django.contrib import admin

from wbportfolio.models import Transaction


@admin.register(Transaction)
class TransactionModelAdmin(admin.ModelAdmin):
    search_fields = ["portfolio__name", "underlying_instrument__computed_str"]
    list_filter = ("transaction_type", "portfolio")
    list_display = (
        "transaction_type",
        "portfolio",
        "underlying_instrument",
        "transaction_date",
        "currency",
        "currency_fx_rate",
        "total_value",
        "total_value_fx_portfolio",
        "import_source",
    )
    fieldsets = (
        (
            "Transaction Information",
            {
                "fields": (
                    ("portfolio", "underlying_instrument", "transaction_type", "import_source", "external_id"),
                    ("transaction_date", "book_date", "value_date"),
                    ("currency", "currency_fx_rate"),
                    ("total_value", "total_value_gross"),
                    ("total_value_fx_portfolio", "total_value_gross_fx_portfolio"),
                    "comment",
                )
            },
        ),
    )
    autocomplete_fields = ["portfolio", "underlying_instrument", "currency"]
    ordering = ("-transaction_date",)
    raw_id_fields = ["import_source"]



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class PortfolioEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return super().get_endpoint()

    def get_update_endpoint(self, **kwargs):
        if self.view.is_portfolio_manager:
            return self.get_endpoint(**kwargs)
        return None

    def get_delete_endpoint(self, **kwargs):
        return self.get_update_endpoint(**kwargs)

    def get_create_endpoint(self, **kwargs):
        return self.get_update_endpoint(**kwargs)


class PortfolioPortfolioThroughModelEndpointConfig(EndpointViewConfig):
    PK_FIELD = "dependency_portfolio"

    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:portfolio-dependencyportfolio-list",
            args=[self.view.kwargs["portfolio_id"]],
            request=self.request,
        )

    def get_instance_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:portfolio-list",
            args=[],
            request=self.request,
        )

    def get_endpoint(self, **kwargs):
        if self.request.user.is_superuser:
            return self.get_list_endpoint(**kwargs)
        return None


class PortfolioTreeGraphChartEndpointConfig(EndpointViewConfig):
    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:portfolio-treegraphchart-list",
            args=[self.view.kwargs["portfolio_id"]],
            request=self.request,
        )


class TopDownPortfolioCompositionPandasEndpointConfig(EndpointViewConfig):
    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:portfolio-topdowncomposition-list",
            args=[self.view.kwargs["portfolio_id"]],
            request=self.request,
        )



# Import necessary modules
from datetime import date, timedelta
from decimal import Decimal
from unittest.mock import call, patch

import pytest
from faker import Faker
from pandas._libs.tslibs.offsets import BDay, BusinessMonthEnd

from wbportfolio.models import Portfolio, RebalancingModel, TradeProposal
from wbportfolio.pms.typing import Portfolio as PortfolioDTO
from wbportfolio.pms.typing import Position

fake = Faker()


# Mark tests to use Django's database
@pytest.mark.django_db
class TestTradeProposal:
    # Test that the checked object is correctly set to the portfolio
    def test_checked_object(self, trade_proposal):
        """
        Verify that the checked object is the portfolio associated with the trade proposal.
        """
        assert trade_proposal.checked_object == trade_proposal.portfolio

    # Test that the evaluation date matches the trade date
    def test_check_evaluation_date(self, trade_proposal):
        """
        Ensure the evaluation date is the same as the trade date.
        """
        assert trade_proposal.check_evaluation_date == trade_proposal.trade_date

    # Test the validated trading service functionality
    def test_validated_trading_service(self, trade_proposal, asset_position_factory, trade_factory):
        """
        Validate that the effective and target portfolios are correctly calculated.
        """
        effective_date = (trade_proposal.trade_date - BDay(1)).date()

        # Create asset positions for testing
        a1 = asset_position_factory.create(
            portfolio=trade_proposal.portfolio, date=effective_date, weighting=Decimal("0.3")
        )
        a2 = asset_position_factory.create(
            portfolio=trade_proposal.portfolio, date=effective_date, weighting=Decimal("0.7")
        )

        # Create trades for testing
        t1 = trade_factory.create(
            trade_proposal=trade_proposal,
            weighting=Decimal("0.05"),
            portfolio=trade_proposal.portfolio,
            transaction_date=trade_proposal.trade_date,
            underlying_instrument=a1.underlying_quote,
        )
        t2 = trade_factory.create(
            trade_proposal=trade_proposal,
            weighting=Decimal("-0.05"),
            portfolio=trade_proposal.portfolio,
            transaction_date=trade_proposal.trade_date,
            underlying_instrument=a2.underlying_quote,
        )

        # Get the validated trading service
        validated_trading_service = trade_proposal.validated_trading_service

        # Assert effective and target portfolios are as expected
        assert validated_trading_service.effective_portfolio.to_dict() == {
            a1.underlying_quote.id: a1.weighting,
            a2.underlying_quote.id: a2.weighting,
        }
        assert validated_trading_service.target_portfolio.to_dict() == {
            a1.underlying_quote.id: a1.weighting + t1.weighting,
            a2.underlying_quote.id: a2.weighting + t2.weighting,
        }

    # Test the calculation of the last effective date
    def test_last_effective_date(self, trade_proposal, asset_position_factory):
        """
        Verify the last effective date is correctly determined based on asset positions.
        """
        # Without any positions, it should be the day before the trade date
        assert (
            trade_proposal.last_effective_date == (trade_proposal.trade_date - BDay(1)).date()
        ), "Last effective date without position should be t-1"

        # Create an asset position before the trade date
        a1 = asset_position_factory.create(
            portfolio=trade_proposal.portfolio, date=(trade_proposal.trade_date - BDay(5)).date()
        )
        a_noise = asset_position_factory.create(portfolio=trade_proposal.portfolio, date=trade_proposal.trade_date)  # noqa

        # The last effective date should still be the day before the trade date due to caching
        assert (
            trade_proposal.last_effective_date == (trade_proposal.trade_date - BDay(1)).date()
        ), "last effective date is cached, so it won't change as is"

        # Reset the cache property to recalculate
        del trade_proposal.last_effective_date

        # Now it should be the date of the latest position before the trade date
        assert (
            trade_proposal.last_effective_date == a1.date
        ), "last effective date is the latest position strictly lower than trade date"

    # Test finding the previous trade proposal
    def test_previous_trade_proposal(self, trade_proposal_factory):
        """
        Ensure the previous trade proposal is correctly identified as the last approved proposal before the current one.
        """
        tp = trade_proposal_factory.create()
        tp_previous_submit = trade_proposal_factory.create(  # noqa
            portfolio=tp.portfolio, status=TradeProposal.Status.SUBMIT, trade_date=(tp.trade_date - BDay(1)).date()
        )
        tp_previous_approve = trade_proposal_factory.create(
            portfolio=tp.portfolio, status=TradeProposal.Status.APPROVED, trade_date=(tp.trade_date - BDay(2)).date()
        )
        tp_next_approve = trade_proposal_factory.create(  # noqa
            portfolio=tp.portfolio, status=TradeProposal.Status.APPROVED, trade_date=(tp.trade_date + BDay(1)).date()
        )

        # The previous valid trade proposal should be the approved one strictly before the current proposal
        assert (
            tp.previous_trade_proposal == tp_previous_approve
        ), "the previous valid trade proposal is the strictly before and approved trade proposal"

    # Test finding the next trade proposal
    def test_next_trade_proposal(self, trade_proposal_factory):
        """
        Verify the next trade proposal is correctly identified as the first approved proposal after the current one.
        """
        tp = trade_proposal_factory.create()
        tp_next_submit = trade_proposal_factory.create(  # noqa
            portfolio=tp.portfolio, status=TradeProposal.Status.SUBMIT, trade_date=(tp.trade_date + BDay(1)).date()
        )
        tp_next_approve = trade_proposal_factory.create(
            portfolio=tp.portfolio, status=TradeProposal.Status.APPROVED, trade_date=(tp.trade_date + BDay(2)).date()
        )
        tp_previous_approve = trade_proposal_factory.create(  # noqa
            portfolio=tp.portfolio, status=TradeProposal.Status.APPROVED, trade_date=(tp.trade_date - BDay(1)).date()
        )

        # The next valid trade proposal should be the approved one strictly after the current proposal
        assert (
            tp.next_trade_proposal == tp_next_approve
        ), "the next valid trade proposal is the strictly after and approved trade proposal"

    # Test getting the default target portfolio
    def test__get_default_target_portfolio(self, trade_proposal, asset_position_factory):
        """
        Ensure the default target portfolio is set to the effective portfolio from the day before the trade date.
        """
        effective_date = (trade_proposal.trade_date - BDay(1)).date()

        # Create asset positions for testing
        a1 = asset_position_factory.create(
            portfolio=trade_proposal.portfolio, date=effective_date, weighting=Decimal("0.3")
        )
        a2 = asset_position_factory.create(
            portfolio=trade_proposal.portfolio, date=effective_date, weighting=Decimal("0.7")
        )
        asset_position_factory.create(portfolio=trade_proposal.portfolio, date=trade_proposal.trade_date)  # noise

        # The default target portfolio should match the effective portfolio
        assert trade_proposal._get_default_target_portfolio().to_dict() == {
            a1.underlying_quote.id: a1.weighting,
            a2.underlying_quote.id: a2.weighting,
        }

    # Test getting the default target portfolio with a rebalancing model
    @patch.object(RebalancingModel, "get_target_portfolio")
    def test__get_default_target_portfolio_with_rebalancer_model(self, mock_fct, trade_proposal, rebalancer_factory):
        """
        Verify that the target portfolio is correctly obtained from a rebalancing model.
        """
        # Expected target portfolio from the rebalancing model
        expected_target_portfolio = PortfolioDTO(
            positions=(Position(underlying_instrument=1, weighting=Decimal(1), date=trade_proposal.trade_date),)
        )
        mock_fct.return_value = expected_target_portfolio

        # Create a rebalancer for testing
        rebalancer = rebalancer_factory.create(
            portfolio=trade_proposal.portfolio, parameters={"rebalancer_parameter": "A"}
        )
        trade_proposal.rebalancing_model = rebalancer.rebalancing_model
        trade_proposal.save()

        # Additional keyword arguments for the rebalancing model
        extra_kwargs = {"test": "test"}

        # Combine rebalancer parameters with extra keyword arguments
        expected_kwargs = rebalancer.parameters
        expected_kwargs.update(extra_kwargs)

        # Assert the target portfolio matches the expected output from the rebalancing model
        assert (
            trade_proposal._get_default_target_portfolio(**extra_kwargs) == expected_target_portfolio
        ), "We expect the target portfolio to be whatever is returned by the rebalancer model"
        mock_fct.assert_called_once_with(
            trade_proposal.portfolio, trade_proposal.trade_date, trade_proposal.last_effective_date, **expected_kwargs
        )

    # Test normalizing trades
    def test_normalize_trades(self, trade_proposal, trade_factory):
        """
        Ensure trades are normalized to sum up to 1, handling quantization errors.
        """
        # Create trades for testing
        t1 = trade_factory.create(
            trade_proposal=trade_proposal,
            transaction_date=trade_proposal.trade_date,
            portfolio=trade_proposal.portfolio,
            weighting=Decimal(0.2),
        )
        t2 = trade_factory.create(
            trade_proposal=trade_proposal,
            transaction_date=trade_proposal.trade_date,
            portfolio=trade_proposal.portfolio,
            weighting=Decimal(0.26),
        )
        t3 = trade_factory.create(
            trade_proposal=trade_proposal,
            transaction_date=trade_proposal.trade_date,
            portfolio=trade_proposal.portfolio,
            weighting=Decimal(0.14),
        )

        # Normalize trades
        trade_proposal.normalize_trades()

        # Refresh trades from the database
        t1.refresh_from_db()
        t2.refresh_from_db()
        t3.refresh_from_db()

        # Expected normalized weights
        normalized_t1_weight = Decimal("0.333333")
        normalized_t2_weight = Decimal("0.433333")
        normalized_t3_weight = Decimal("0.233333")

        # Calculate quantization error
        quantize_error = Decimal(1) - (normalized_t1_weight + normalized_t2_weight + normalized_t3_weight)

        # Assert quantization error exists and weights are normalized correctly
        assert quantize_error
        assert t1.weighting == normalized_t1_weight
        assert t2.weighting == normalized_t2_weight + quantize_error  # Add quantize error to the largest position
        assert t3.weighting == normalized_t3_weight

    # Test resetting trades
    def test_reset_trades(self, trade_proposal, instrument_factory, asset_position_factory):
        """
        Verify trades are correctly reset based on effective and target portfolios.
        """
        effective_date = trade_proposal.last_effective_date

        # Create instruments for testing
        i1 = instrument_factory.create(currency=trade_proposal.portfolio.currency)
        i2 = instrument_factory.create(currency=trade_proposal.portfolio.currency)
        i3 = instrument_factory.create(currency=trade_proposal.portfolio.currency)

        # Build initial effective portfolio constituting only from two positions of i1 and i2
        asset_position_factory.create(
            portfolio=trade_proposal.portfolio, date=effective_date, underlying_instrument=i1, weighting=Decimal("0.7")
        )
        asset_position_factory.create(
            portfolio=trade_proposal.portfolio, date=effective_date, underlying_instrument=i2, weighting=Decimal("0.3")
        )

        # build the target portfolio
        target_portfolio = PortfolioDTO(
            positions=(
                Position(underlying_instrument=i2.id, date=trade_proposal.trade_date, weighting=Decimal("0.4")),
                Position(underlying_instrument=i3.id, date=trade_proposal.trade_date, weighting=Decimal("0.6")),
            )
        )

        # Reset trades
        trade_proposal.reset_trades(target_portfolio=target_portfolio)

        # Get trades for each instrument
        t1 = trade_proposal.trades.get(underlying_instrument=i1)
        t2 = trade_proposal.trades.get(underlying_instrument=i2)
        t3 = trade_proposal.trades.get(underlying_instrument=i3)

        # Assert trade weights are correctly reset
        assert t1.weighting == Decimal("-0.7")
        assert t2.weighting == Decimal("0.1")
        assert t3.weighting == Decimal("0.6")

        # build the target portfolio
        new_target_portfolio = PortfolioDTO(
            positions=(
                Position(underlying_instrument=i1.id, date=trade_proposal.trade_date, weighting=Decimal("0.2")),
                Position(underlying_instrument=i2.id, date=trade_proposal.trade_date, weighting=Decimal("0.3")),
                Position(underlying_instrument=i3.id, date=trade_proposal.trade_date, weighting=Decimal("0.5")),
            )
        )

        trade_proposal.reset_trades(target_portfolio=new_target_portfolio)
        # Refetch the trades for each instrument
        t1.refresh_from_db()
        t2.refresh_from_db()
        t3.refresh_from_db()
        # Assert existing trade weights are correctly updated
        assert t1.weighting == Decimal("-0.5")
        assert t2.weighting == Decimal("0")
        assert t3.weighting == Decimal("0.5")

    # Test replaying trade proposals
    @patch.object(Portfolio, "batch_portfolio")
    def test_replay(self, mock_fct, trade_proposal_factory):
        """
        Ensure replaying trade proposals correctly calls batch_portfolio for each period.
        """
        mock_fct.return_value = None

        # Create approved trade proposals for testing
        tp0 = trade_proposal_factory.create(status=TradeProposal.Status.APPROVED)
        tp1 = trade_proposal_factory.create(
            portfolio=tp0.portfolio,
            status=TradeProposal.Status.APPROVED,
            trade_date=(tp0.trade_date + BusinessMonthEnd(1)).date(),
        )
        tp2 = trade_proposal_factory.create(
            portfolio=tp0.portfolio,
            status=TradeProposal.Status.APPROVED,
            trade_date=(tp1.trade_date + BusinessMonthEnd(1)).date(),
        )

        # Replay trade proposals
        tp0.replay()

        # Expected calls to batch_portfolio
        expected_calls = [
            call(tp0.trade_date, tp1.trade_date - timedelta(days=1)),
            call(tp1.trade_date, tp2.trade_date - timedelta(days=1)),
            call(tp2.trade_date, date.today()),
        ]

        # Assert batch_portfolio was called as expected
        mock_fct.assert_has_calls(expected_calls)

        # Test stopping replay on a non-approved proposal
        tp1.status = TradeProposal.Status.FAILED
        tp1.save()
        expected_calls = [call(tp0.trade_date, tp1.trade_date - timedelta(days=1))]
        mock_fct.assert_has_calls(expected_calls)

    # Test estimating shares for a trade
    @patch.object(Portfolio, "get_total_asset_value")
    def test_get_estimated_shares(
        self, mock_fct, trade_proposal, trade_factory, instrument_price_factory, instrument_factory
    ):
        """
        Verify shares estimation based on trade weighting and instrument price.
        """
        portfolio = trade_proposal.portfolio
        instrument = instrument_factory.create(currency=portfolio.currency)
        trade = trade_factory.create(
            trade_proposal=trade_proposal,
            transaction_date=trade_proposal.trade_date,
            portfolio=portfolio,
            underlying_instrument=instrument,
        )
        trade.refresh_from_db()
        underlying_quote_price = instrument_price_factory.create(instrument=instrument, date=trade.transaction_date)
        mock_fct.return_value = Decimal(1_000_000)  # 1 million cash

        # Assert estimated shares are correctly calculated
        assert (
            trade_proposal.get_estimated_shares(trade.weighting, trade.underlying_instrument)
            == Decimal(1_000_000) * trade.weighting / underlying_quote_price.net_value
        )

    @patch.object(Portfolio, "get_total_asset_value")
    def test_get_estimated_target_cash(self, mock_fct, trade_proposal, trade_factory, cash_factory):
        mock_fct.return_value = Decimal(1_000_000)  # 1 million cash
        cash = cash_factory.create(currency=trade_proposal.portfolio.currency)
        trade_factory.create(  # equity trade
            trade_proposal=trade_proposal,
            transaction_date=trade_proposal.trade_date,
            portfolio=trade_proposal.portfolio,
            weighting=Decimal("0.7"),
        )
        trade_factory.create(  # cash trade
            trade_proposal=trade_proposal,
            transaction_date=trade_proposal.trade_date,
            portfolio=trade_proposal.portfolio,
            underlying_instrument=cash,
            weighting=Decimal("0.2"),
        )

        target_cash_position = trade_proposal.get_estimated_target_cash(trade_proposal.portfolio.currency)
        assert target_cash_position.weighting == Decimal("0.2") + Decimal("1.0") - (Decimal("0.7") + Decimal("0.2"))
        assert target_cash_position.initial_shares == Decimal(1_000_000) * Decimal("0.3")



from .adjustments import AdjustmentFactory
from .assets import AssetPositionFactory
from .claim import (
    ApprovedClaimFactory,
    ClaimFactory,
    NegativeClaimFactory,
    PositiveClaimFactory,
)
from .custodians import CustodianFactory
from .dividends import DividendTransactionsFactory
from .fees import FeesFactory
from .portfolios import (
    InstrumentPortfolioThroughModelFactory,
    ModelPortfolioFactory,
    PortfolioFactory,
)
from .portfolio_swing_pricings import PortfolioSwingPricingFactory
from .portfolio_cash_targets import PortfolioCashTargetFactory
from .portfolio_cash_flow import DailyPortfolioCashFlowFactory
from .product_groups import ProductGroupFactory, ProductGroupRepresentantFactory
from .products import IndexProductFactory, ProductFactory, WhiteLabelProductFactory, ModelPortfolioWithBaseProductFactory
from .reconciliations import AccountReconciliationFactory, AccountReconciliationLineFactory
from .roles import ManagerPortfolioRoleFactory, ProductPortfolioRoleFactory
from .trades import CustomerTradeFactory, TradeFactory, TradeProposalFactory
from .transactions import TransactionFactory
from .indexes import IndexFactory
from .rebalancing import (RebalancingModelFactory, RebalancerFactory)



import re
from typing import Any, Dict

from wbportfolio.models import Product, Register, Trade


def get_portfolio_id(x):
    product = Product.objects.get(id=x)
    portfolio = product.primary_portfolio
    return portfolio.id


def parse_transaction_reference(trans_ref: str) -> tuple[str, str, bool]:
    pattern = r"([0-9]{7})([0-9]{8})([CD]{1})"
    transaction_id, outlet_id, credit_debit = re.findall(pattern, trans_ref)[0]
    return transaction_id, outlet_id, credit_debit == "C"


def assemble_transaction_reference(data: Dict[str, Any]) -> str:
    transaction_ref = data["external_identifier2"]
    register_reference = data["register__register_reference"]
    if data.get("TRANSFER_REGISTER", None):
        register_reference = data["TRANSFER_REGISTER"]

    outlet = Register.objects.get(register_reference=register_reference).outlet_reference
    credit_debit = "C" if data["shares"] > 0 else "D"

    return f"{int(transaction_ref):07}{int(outlet):08}{credit_debit}"


def create_transaction_reference(customer_trade: "Trade") -> str:
    transaction_id = customer_trade.external_identifier2
    outlet_id = customer_trade.register.clearing_reference
    credit_debit = "C" if customer_trade.initial_shares > 0 else "D"

    return f"{transaction_id:07}{outlet_id:08}{credit_debit}"



import random
from datetime import timedelta

import pytest
from django.db.models import ProtectedError
from faker import Faker

from wbportfolio.models import Trade
from wbportfolio.models.transactions.claim import Claim

fake = Faker()


@pytest.mark.django_db
class TestCustomerTradeModel:
    def test_delete_without_claims(self, trade):
        """
        Simple test to check if a trade without claim is properly deleted
        """
        trade.delete()
        with pytest.raises(Trade.DoesNotExist):
            trade.refresh_from_db()

    @pytest.mark.parametrize("claim_status", Claim.Status.names)
    def test_delete_trade_with_claim(self, customer_trade_factory, claim_factory, claim_status):
        """
        Simple test to check if a trade with claim a valid claim (Pending or Approved9 can't be deleted but
        a trade with a draft, withdrawn or auto approved claim can.
        """
        customer_trade = customer_trade_factory.create()
        claim = claim_factory.create(trade=customer_trade, status=claim_status)

        # If claim is among the non approved type, we expect the deletion to succeed and the claim to be unlinked
        if claim.status in [Claim.Status.DRAFT, Claim.Status.AUTO_MATCHED, Claim.Status.WITHDRAWN]:
            customer_trade.delete()
            claim.refresh_from_db()
            assert not claim.trade
            with pytest.raises(Trade.DoesNotExist):
                customer_trade.refresh_from_db()
        else:
            # If the claim is approved or pending, and in absence of any other similar trades to reassign the claim to, we expect the deletion to failed
            with pytest.raises(ProtectedError):
                customer_trade.delete()

    def test_delete_trade_with_approved_claims_but_similar_trades_summing_0(
        self, customer_trade_factory, claim_factory
    ):
        """
        A test to check that even trade with approved claims attach can be deleted if the sum of the shares of the similar trades the same date equals to 0
        """
        trade1 = customer_trade_factory.create()
        claim1 = claim_factory.create(trade=trade1, status=Claim.Status.APPROVED)
        trade1.marked_for_deletion = True
        trade1.save()

        # Claims is attached, no other similar trade is present, we don't expect the deletion to be possible
        with pytest.raises(ProtectedError):
            trade1.delete()

        # We create a negative trade that will annihilate trade1
        trade2 = customer_trade_factory.create(
            underlying_instrument=trade1.underlying_instrument,
            portfolio=trade1.portfolio,
            transaction_date=trade1.transaction_date,
            shares=-trade1.shares,
        )
        # That we claim
        claim2 = claim_factory.create(trade=trade2, status=Claim.Status.APPROVED)
        trade2.marked_for_deletion = True
        trade2.save()

        # Now, sum of all trades linked to this product at that date sums to 0, we can destroy all trades and unliked the associated claims
        trade1.delete()
        claim1.refresh_from_db()
        claim2.refresh_from_db()
        assert not claim1.trade
        assert claim1.status == Claim.Status.DRAFT
        assert not claim2.trade
        assert claim2.status == Claim.Status.DRAFT

        # trade2 is unclaimed and can be safely deleted
        trade2.delete()
        with pytest.raises(Trade.DoesNotExist):
            trade2.refresh_from_db()

    @pytest.mark.parametrize(
        "allowed_timedeltda", [random.randint(-Trade.TRADE_WINDOW_INTERVAL, Trade.TRADE_WINDOW_INTERVAL)]
    )
    def test_trade_delete_shift_claims(self, customer_trade_factory, claim_factory, allowed_timedeltda):
        """
        Test main functionality: when a trade is linked to an approved claim and is marked for deletion, we try to find the closest trade non marked for deletion or pending and unclaimed.

        If this succeed, the claim is forwarded to this other trade. Otherwise, it will fail.
        """
        trade1 = customer_trade_factory.create()
        claim1 = claim_factory.create(trade=trade1, status=Claim.Status.APPROVED)
        trade1.marked_for_deletion = True
        trade1.save()

        # We expect the claim to be reassigned to this similar "non marked for deletion" trade
        similar_unclaimed_trade = customer_trade_factory.create(
            marked_for_deletion=False,
            pending=False,
            underlying_instrument=trade1.underlying_instrument,
            portfolio=trade1.portfolio,
            transaction_subtype=trade1.transaction_subtype,
            transaction_date=trade1.transaction_date + timedelta(days=allowed_timedeltda),
            shares=trade1.shares,
        )
        trade1.delete()
        with pytest.raises(Trade.DoesNotExist):
            trade1.refresh_from_db()
        claim1.refresh_from_db()
        assert claim1.trade == similar_unclaimed_trade



import codecs
import csv
import datetime
import re

from wbportfolio.import_export.utils import convert_string_to_number
from wbportfolio.models import Product, Trade

product_mapping = {
    "2304": {
        "A": "CH0442615701",
        "B": "CH0442770316",
    },
    "2316": {
        "A": "CH0583763534",
        "B": "CH0583763542",
    },
}


def file_name_parse(file_name):
    identifier = re.findall("([0-9]{4}).*", file_name)
    assert len(identifier) == 1, "Not exactly one identifier was found."
    return identifier[0]


def parse(import_source):
    # Get the identifier from the file name to know which product group it is
    identifier = file_name_parse(import_source.file.name)

    # Load file into a CSV DictReader
    csv_file = import_source.file.open()
    csv_file_type = type(csv_file.read())

    # If the file is a byte string, then we need to convert it.
    if csv_file_type is bytes:
        csv_file = codecs.iterdecode(csv_file, "latin1")

    # Read file into a CSV Dict Reader
    csv_reader = csv.DictReader(csv_file, delimiter=";")

    # Iterate through the CSV File and parse the data into a list
    data = list()
    for customer_trade in csv_reader:
        product = Product.objects.get(isin=product_mapping[identifier][customer_trade["Anteilklasse"].strip()])

        transaction_date = datetime.datetime.strptime(customer_trade["Datum"], "%Y%m%d").date()
        shares = round(convert_string_to_number(customer_trade["Saldo - Anzahl"]), 4)
        portfolio = product.primary_portfolio
        data.append(
            {
                "underlying_instrument": {"id": product.id, "instrument_type": "product"},
                "portfolio": portfolio.id,
                "currency__key": product.currency.key,
                "transaction_date": transaction_date.strftime("%Y-%m-%d"),
                "value_date": transaction_date.strftime("%Y-%m-%d"),
                "shares": shares,
                "bank": "Tellco Trade",
                "transaction_subtype": Trade.Type.REDEMPTION if shares < 0 else Trade.Type.SUBSCRIPTION,
                "price": round(convert_string_to_number(customer_trade["NIW"]), 4),
            }
        )
    csv_file.close()
    return {"data": data}



# Generated by Django 5.0.6 on 2024-07-19 14:20

from decimal import Decimal

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbcrm", "0009_alter_account_is_public"),
        ("wbportfolio", "0062_alter_dailyportfoliocashflow_options"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="AccountReconciliation",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("reconciliation_date", models.DateField()),
                ("approved_dt", models.DateTimeField(blank=True, null=True, verbose_name="Approved Timestamp")),
                (
                    "account",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="wbportfolio_reconciliations",
                        to="wbcrm.account",
                    ),
                ),
                (
                    "approved_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="Approved By",
                    ),
                ),
                (
                    "creator",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT, related_name="+", to=settings.AUTH_USER_MODEL
                    ),
                ),
            ],
            options={
                "verbose_name": "Account Holdings Reconciliation",
                "verbose_name_plural": "Account Holdings Reconciliations",
                "notification_types": [
                    (
                        "wbportfolio.account_reconciliation.notify",
                        "Account Reconciliation Events",
                        "A notification that informs the user about an account reconciliation event.",
                        True,
                        True,
                        True,
                    )
                ],
            },
        ),
        migrations.CreateModel(
            name="AccountReconciliationLine",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "price",
                    models.DecimalField(
                        decimal_places=4,
                        default=Decimal("0"),
                        help_text="The last share price of the product",
                        max_digits=18,
                    ),
                ),
                ("price_date", models.DateField()),
                (
                    "shares",
                    models.IntegerField(default=0, help_text="The number of shares computed through the Workbench"),
                ),
                (
                    "nominal_value",
                    models.IntegerField(default=0, help_text="The nominal value computed through the Workbench"),
                ),
                (
                    "shares_external",
                    models.IntegerField(
                        default=0,
                        help_text="The number of shares externally provided through the Account holder. Initially set to the number of shares computed through the Workbench",
                    ),
                ),
                (
                    "nominal_value_external",
                    models.IntegerField(
                        default=0,
                        help_text="The nominal value externally provided through the Account holder. Initially set to the number of shares computed through the Workbench",
                    ),
                ),
                (
                    "product",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="account_reconciliation_lines",
                        to="wbportfolio.product",
                    ),
                ),
                (
                    "reconciliation",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="lines",
                        to="wbportfolio.accountreconciliation",
                    ),
                ),
            ],
            options={
                "verbose_name": "Account Reconciliation (Product)",
                "verbose_name_plural": "Account Reconciliation (Products)",
            },
        ),
        migrations.AddConstraint(
            model_name="accountreconciliation",
            constraint=models.UniqueConstraint(fields=("reconciliation_date", "account"), name="unique_date_account"),
        ),
        migrations.AddConstraint(
            model_name="accountreconciliationline",
            constraint=models.UniqueConstraint(
                fields=("reconciliation", "product"), name="unique_reconcilation_product"
            ),
        ),
    ]



from datetime import date, timedelta
from typing import Generator

import pandas as pd
from django.contrib.contenttypes.models import ContentType
from django.db import models
from pandas.tseries.offsets import BDay
from wbcompliance.models.risk_management import backend
from wbcore import serializers as wb_serializers
from wbfdm.backends.dto import PriceDTO
from wbfdm.models import Instrument, InstrumentPrice, InstrumentType
from wbfdm.serializers import (
    InstrumentTypeRepresentationSerializer,
    SecurityRepresentationSerializer,
)

from wbportfolio.models import InstrumentPortfolioThroughModel, Portfolio


class ActivePortfolioRelationshipMixin(backend.AbstractRuleBackend):
    OBJECT_FIELD_NAME: str = "portfolio"

    portfolio: Portfolio

    def is_passive_evaluation_valid(self) -> bool:
        return self.portfolio.imported_assets.filter(date=self.evaluation_date).exists()

    @classmethod
    def get_allowed_content_type(cls) -> "ContentType":
        return ContentType.objects.get_for_model(Portfolio)

    def _build_dto_args(self):
        return (self.portfolio._build_dto(self.evaluation_date, is_estimated=False),)

    @classmethod
    def get_all_active_relationships(cls) -> models.QuerySet:
        valid_relationships = InstrumentPortfolioThroughModel.objects.filter(
            instrument__instrument_type__key="product"
        ).values("portfolio")
        return Portfolio.objects.filter(id__in=valid_relationships, is_tracked=True)


class ActiveProductRelationshipMixin(backend.AbstractRuleBackend):
    OBJECT_FIELD_NAME: str = "product"

    @classmethod
    def get_allowed_content_type(cls) -> "ContentType":
        return ContentType.objects.get_for_model(Instrument)

    @classmethod
    def get_all_active_relationships(cls) -> models.QuerySet:
        return Instrument.active_objects.filter(instrument_type__key="product")


class StopLossMixin(backend.AbstractRuleBackend):
    class FreqChoices(models.TextChoices):
        BUSINESS_DAY = "B", "Business Day"
        WEEKLY_FRIDAY = "W-FRI", "Friday to Friday"
        BUSINESS_MONTHLY = "BME", "Business Monthly"
        BUSINESS_YEARLY = "BYE", "Business Yearly"

    class DateIntervalOption(models.TextChoices):
        ROLLING_WINDOWS = "ROLLING_WINDOWS", "Rolling Window"
        FREQUENCY = "FREQUENCY", "Frequency"

    class DynamicBenchmarkType(models.TextChoices):
        PORTFOLIO = "PORTFOLIO", "Primary Portfolio"
        PRIMARY_BENCHMARK = "PRIMARY_BENCHMARK", "Primary Benchmark"

    class FieldType(models.TextChoices):
        OUTSTANDING_SHARES = "outstanding_shares", "Outstanding shares"
        NET_VALUE = "net_value", "Net Value"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.portfolio = self.product.portfolio

    @classmethod
    def get_serializer_class(cls) -> wb_serializers.Serializer:
        class RuleBackendSerializer(wb_serializers.Serializer):
            freq = wb_serializers.ChoiceField(
                choices=cls.FreqChoices,
                default=cls.FreqChoices.WEEKLY_FRIDAY,
                label="Frequency",
                help_text="Valid only if the interval Option is Frequency. Specify the frequency use for aggregation",
            )
            date_interval_option = wb_serializers.ChoiceField(
                choices=cls.DateIntervalOption,
                default=cls.DateIntervalOption.ROLLING_WINDOWS,
                label="Interval Option",
            )
            dynamic_benchmark_type = wb_serializers.ChoiceField(
                choices=cls.DynamicBenchmarkType,
                default=None,
                allow_null=True,
                label="Benchmark Type",
                help_text="If specified, will compare the stop loss against the instrument related potential benchmark performance",
            )
            rolling_window_interval = wb_serializers.IntegerField(
                default=7,
                label="Rolling Window Interval",
                help_text="Valid only if interval option is Rolling Window. Specify the number of day for sampling",
            )
            static_benchmark = wb_serializers.PrimaryKeyRelatedField(
                queryset=Instrument.objects.all(),
                default=None,
                allow_null=True,
                label="Static Benchmark (If any)",
                help_text="If specified, will compare the stop loss against this benchmark",
            )
            asset_class = wb_serializers.PrimaryKeyRelatedField(
                queryset=InstrumentType.objects.all(),
                default=None,
                allow_null=True,
                label="Only Asset Class",
            )
            _asset_class = InstrumentTypeRepresentationSerializer(source="asset_class")

            _static_benchmark = SecurityRepresentationSerializer(
                source="static_benchmark", default=None, allow_null=True
            )
            field = wb_serializers.ChoiceField(
                choices=cls.FieldType.choices,
                default=cls.FieldType.NET_VALUE,
                label="Field",
            )
            penny_stock_max_abs_net_value = wb_serializers.FloatField(
                default=0.001, label="Penny Stock Max Absolute Net value"
            )

            @classmethod
            def get_parameter_fields(cls):
                return [
                    "freq",
                    "date_interval_option",
                    "dynamic_benchmark_type",
                    "rolling_window_interval",
                    "static_benchmark",
                    "asset_class",
                    "field",
                    "penny_stock_max_abs_net_value",
                ]

        return RuleBackendSerializer

    @property
    def benchmark(self):
        if self.dynamic_benchmark_type == self.DynamicBenchmarkType.PRIMARY_BENCHMARK.name:
            return self.product.primary_benchmark
        elif self.dynamic_benchmark_type == self.DynamicBenchmarkType.PORTFOLIO.name:
            return self.product
        elif self.static_benchmark:
            return self.static_benchmark

    def is_passive_evaluation_valid(self) -> bool:
        try:
            last_price = self.product.get_price(self.evaluation_date)
            base_condition = last_price != 0
            if (benchmark := self.benchmark) and (last_benchmark_price := benchmark.get_price(self.evaluation_date)):
                return (last_benchmark_price != 0) and base_condition
            return base_condition
        except ValueError:
            return False

    def _generate_incidents(
        self,
        tested_instrument_id: int,
        perf_instrument: float,
        perf_benchmark: float,
    ) -> Generator[backend.IncidentResult, None, None]:
        total_perf = perf_instrument if perf_benchmark is None else perf_instrument - perf_benchmark
        field_label = self.FieldType(self.field).label
        for threshold in self.thresholds:
            if threshold.is_inrange(total_perf):
                instrument = Instrument.objects.get(id=tested_instrument_id)
                report_details = {
                    "Absolute Percentage": f"{perf_instrument:,.3%}",
                    "Field": field_label,
                }
                if self.benchmark:
                    report_details[f"Relative Percentage VS {str(self.benchmark)}"] = f"{total_perf:,.3%}"
                if total_perf < 0:
                    breached_value = f'<span style="color:red">{total_perf:+,.2%}</span>'
                else:
                    breached_value = f'<span style="color:green">{total_perf:+,.2%}</span>'
                yield backend.IncidentResult(
                    breached_object=instrument,
                    breached_object_repr=str(instrument),
                    breached_value=breached_value,
                    report_details=report_details,
                    severity=threshold.severity,
                )

    def _get_start_interval(self) -> date:
        if self.date_interval_option == self.DateIntervalOption.FREQUENCY.name:
            return pd.date_range(end=self.evaluation_date, periods=2, freq=self.FreqChoices(self.freq).value)[0].date()
        else:
            return (self.evaluation_date - timedelta(days=self.rolling_window_interval - 1) - BDay(1)).date()

    def _get_performance(self, valuation_dto: PriceDTO) -> float:
        if not valuation_dto:
            return 0.0
        start_date = self._get_start_interval()
        if (
            qs := InstrumentPrice.objects.filter(
                calculated=False,
                instrument_id=valuation_dto.instrument,
                date__lte=start_date,
                date__gte=start_date - BDay(2),  # we allow 2 business day interval in case of market holiday
            ).exclude(
                net_value__lte=self.penny_stock_max_abs_net_value, net_value__gte=-self.penny_stock_max_abs_net_value
            )
        ).exists():
            last_value = float(getattr(qs.latest("date"), self.field))
            return float(getattr(valuation_dto, self.field)) / last_value - 1
        return 0.0

    def _build_dto_args(self) -> tuple[PriceDTO, PriceDTO | None]:
        if benchmark := self.benchmark:
            return self.product._build_dto(self.evaluation_date), benchmark._build_dto(self.evaluation_date)
        return self.product._build_dto(self.evaluation_date), None



from io import BytesIO
from typing import Dict

import numpy as np
import pandas as pd
from wbcore.contrib.io.models import ImportSource

from wbportfolio.models import Product, Trade


def parse_row(obj: Dict, import_source: ImportSource) -> Dict:
    try:
        isin = obj["underlying_instrument__isin"]
        product = Product.objects.get(isin=isin)
        shares = obj["shares"]
        if import_source.source.import_parameters.get("negate_shares", False):
            shares = -1 * shares
        portfolio = product.primary_portfolio
        return {
            "underlying_instrument": {"id": product.id, "instrument_type": "product"},
            "currency__key": product.currency.key,
            "portfolio": portfolio.id,
            "transaction_date": obj["transaction_date"].strftime("%Y-%m-%d"),
            "shares": shares,
            "bank": obj["custodian"],
            "transaction_subtype": Trade.Type.REDEMPTION if shares < 0 else Trade.Type.SUBSCRIPTION,
            "price": round(obj["price"], 6),
        }
    except Product.DoesNotExist:
        import_source.log += f"Product with ISIN {isin} does not exists."


def parse(import_source):
    df = pd.read_excel(BytesIO(import_source.file.read()), engine="openpyxl")
    if "Trade Date" not in df.columns:
        xx, yy = np.where(df == "Trade Date")
        df = df.iloc[xx[0] :, yy[0] :]
        df = df.rename(columns=df.iloc[0]).drop(df.index[0]).dropna(how="all")
    df = df.rename(columns=lambda x: x.lower())
    df = df.rename(
        columns={
            "trade date": "transaction_date",
            "isin": "underlying_instrument__isin",
            "client side": "client_side",
            "net quantity": "shares",
            "price": "price",
            "custodian": "custodian",
            "name": "custodian",
            "buy/sell": "client_side",
            "nominal": "shares",
        }
    )
    df["transaction_date"] = pd.to_datetime(df["transaction_date"])
    df["underlying_instrument__isin"] = df["underlying_instrument__isin"].str.strip()
    df = df.replace([np.inf, -np.inf, np.nan], None)
    df.loc[df["custodian"].isnull(), "custodian"] = "N/A"
    data = list()
    for d in df.to_dict("records"):
        data.append(parse_row(d, import_source))

    return {"data": data}



from wbcore import serializers as wb_serializers
from wbcore.contrib.directory.serializers import PersonRepresentationSerializer
from wbfdm.serializers import SecurityRepresentationSerializer

from wbportfolio.models import Adjustment


class AdjustmentModelSerializer(wb_serializers.ModelSerializer):
    _instrument = SecurityRepresentationSerializer(source="instrument")
    _last_handler = PersonRepresentationSerializer(source="last_handler")

    class Meta:
        model = Adjustment
        fields = (
            "id",
            "date",
            "instrument",
            "_instrument",
            "status",
            "factor",
            "cumulative_factor",
            "_last_handler",
            "last_handler",
            "_additional_resources",
        )



from contextlib import suppress
from datetime import date, timedelta

from django.db.models import ProtectedError, Q

from wbportfolio.models import Portfolio, Trade
from wbportfolio.models.products import Product, update_outstanding_shares_as_task

from .fdm.tasks import *  # noqa


@shared_task(queue="portfolio")
def daily_active_product_task(today: date | None = None):
    if not today:
        today = date.today()
    qs = Product.active_objects.all()
    for product in tqdm(qs, total=qs.count()):
        update_outstanding_shares_as_task(product.id)
        product.check_and_notify_product_termination_on_date(today)


@shared_task(queue="portfolio")
def periodically_clean_marked_for_deletion_trades(max_allowed_iterations: int = 5):
    # Get all trade marked for deletion or pending and older than 7 days (i.e. After 7 days, we consider the pending trade obselete)
    qs = Trade.objects.filter(
        Q(marked_for_deletion=True) | (Q(pending=True) & Q(transaction_date__lt=date.today() - timedelta(days=7)))
    )
    i = 0

    # We try several times in case the trades deletion mechanism shifts the marked for deletion tag forwards
    while i < max_allowed_iterations and qs.exists():
        for t in qs:
            with suppress(ProtectedError):
                t.delete()
        qs = Trade.objects.filter(marked_for_deletion=True)
        i += 1


# A Task to run every day to update automatically the preferred classification
# per instrument of each wbportfolio containing assets.
@shared_task(queue="portfolio")
def update_preferred_classification_per_instrument_and_portfolio_as_task():
    for portfolio in Portfolio.tracked_objects.all():
        portfolio.update_preferred_classification_per_instrument()


# This task needs to run at fix interval. It will trigger the basic wbportfolio synchronization update:
# - Fetch for stainly price at t-1
# - propagate (or update) t-2 asset positions into t-1
# - Synchronize wbportfolio at t-1
# - Compute Instrument Price estimate at t-1



import datetime as dt

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from django.db.models import QuerySet
from django.shortcuts import get_object_or_404
from django.utils.functional import cached_property
from wbcore import viewsets
from wbcore.contrib.io.viewsets import ExportPandasAPIViewSet
from wbcore.filters import DjangoFilterBackend
from wbcore.pandas import fields as pf
from wbfdm.models import (
    ClassificationGroup,
    Instrument,
    InstrumentClassificationThroughModel,
    InstrumentType,
)

from wbportfolio.filters.assets import DistributionFilter
from wbportfolio.models import (
    AssetPosition,
    AssetPositionGroupBy,
    Portfolio,
    PortfolioRole,
)

from ..configs.buttons.assets import (
    DistributionChartButtonConfig,
    DistributionTableButtonConfig,
)
from ..configs.display.assets import DistributionTableDisplayConfig
from ..configs.endpoints.assets import (
    DistributionChartEndpointConfig,
    DistributionTableEndpointConfig,
)
from ..configs.titles.assets import (
    DistributionChartTitleConfig,
    DistributionTableTitleConfig,
)


class AbstractDistributionMixin:
    AUTORESIZE = False
    queryset = AssetPosition.objects.all()
    filterset_class = DistributionFilter
    filter_backends = (DjangoFilterBackend,)

    @cached_property
    def classification_group(self):
        try:
            return ClassificationGroup.objects.get(id=self.request.GET.get("group_by_classification_group"))
        except ClassificationGroup.DoesNotExist:
            return ClassificationGroup.objects.get(is_primary=True)

    @cached_property
    def classification_field_names(self):
        return [f"classification__{field_name}__name" for field_name in self.classification_group.get_fields_names()]

    @cached_property
    def classification_levels_representation(self):
        return self.classification_group.get_levels_representation()

    @cached_property
    def classification_columns_map(self):
        return dict(
            zip(["classification__name", *self.classification_field_names], self.classification_levels_representation)
        )

    def _generate_classification_df(self, queryset):
        df = pd.DataFrame(
            queryset.filter(underlying_instrument__instrument_type=InstrumentType.EQUITY).values(
                "weighting", "underlying_instrument"
            ),
            columns=["weighting", "underlying_instrument"],
        )
        df.underlying_instrument = df.underlying_instrument.map(
            dict(
                Instrument.objects.filter(id__in=df.underlying_instrument)
                .annotate_base_data()
                .values_list("id", "root")
            )
        )
        df = df.groupby("underlying_instrument").sum()
        classifications = InstrumentClassificationThroughModel.objects.filter(
            classification__group=self.classification_group, instrument__in=df.index
        )
        df_classification = pd.DataFrame(
            classifications.values(
                "instrument",
                "classification__name",
                *self.classification_field_names,
            )
        )
        if df_classification.empty:
            return pd.DataFrame()
        return pd.concat([df, df_classification.groupby("instrument").first()], axis=1).replace(
            [np.inf, -np.inf, np.nan], "N/A"
        )

    def get_queryset(self):
        portfolio = get_object_or_404(Portfolio, id=self.kwargs["portfolio_id"])
        if (
            PortfolioRole.is_analyst(self.request.user.profile, portfolio=portfolio)
            or self.request.user.profile.is_internal
        ):
            return super().get_queryset().filter(portfolio=portfolio)
        return AssetPosition.objects.none()

    @staticmethod
    def dataframe_group_by_instrument(df: pd.DataFrame) -> pd.DataFrame:
        if df.empty:
            return pd.DataFrame()
        return df.groupby("aggregated_title").sum().sort_values(by="weighting", ascending=False)

    def dataframe_groupby_with_class_method(self, qs: QuerySet, class_method: classmethod):
        df = pd.DataFrame()
        if qs.exists():
            df = self.dataframe_group_by_instrument(
                pd.DataFrame(class_method(qs).values("weighting", "aggregated_title"))
            )
        return df


class DistributionChartViewSet(AbstractDistributionMixin, viewsets.ChartViewSet):
    title_config_class = DistributionChartTitleConfig
    endpoint_config_class = DistributionChartEndpointConfig
    button_config_class = DistributionChartButtonConfig

    @staticmethod
    def pie_chart(df: pd.DataFrame) -> go.Figure:
        fig = go.Figure()
        if not df.empty:
            fig.add_pie(
                labels=df.index,
                values=df.weighting.mul(100),
                marker=dict(colors=px.colors.qualitative.Plotly[: df.shape[0]], line=dict(color="#000000", width=2)),
                hovertemplate="<b>%{label}</b><extra></extra>",
            )
        return fig

    def get_plotly(self, queryset):
        fig = go.Figure()
        group_by = self.request.GET.get("group_by", "COUNTRY")
        class_method = AssetPositionGroupBy.get_class_method_group_by(name=group_by)
        queryset_without_cash = queryset.exclude(underlying_instrument__is_cash=True)
        if group_by not in ["INDUSTRY", "CURRENCY"]:
            df = self.dataframe_groupby_with_class_method(qs=queryset_without_cash, class_method=class_method)
            fig = self.pie_chart(df=df)
        elif group_by == "CURRENCY":
            df = self.dataframe_groupby_with_class_method(qs=queryset, class_method=class_method)
            fig = self.pie_chart(df=df)
        else:
            df = self._generate_classification_df(queryset_without_cash)
            if not df.empty:
                df["weighting"] = df.weighting / df.weighting.sum()
                df.weighting = df.weighting.astype("float")
                df = df.reset_index().rename(
                    columns={**self.classification_columns_map, "weighting": "weight", "index": "Equity"}
                )

                levels = [*self.classification_levels_representation[::-1], "Equity"]
                df["Equity"] = df["Equity"].map(
                    dict(Instrument.objects.filter(id__in=df["Equity"]).values_list("id", "name_repr"))
                )
                portfolio = Portfolio.objects.get(id=self.kwargs["portfolio_id"])
                if not PortfolioRole.is_analyst(self.request.user.profile, portfolio=portfolio):
                    del df["Equity"]
                    levels.remove("Equity")
                fig = px.sunburst(
                    df,
                    path=levels,
                    values="weight",
                    hover_data={"weight": ":.2%"},
                )
                fig.update_traces(hovertemplate="<b>%{label}</b><br>Weight = %{customdata:.3p}")
        return fig


class DistributionTableViewSet(AbstractDistributionMixin, ExportPandasAPIViewSet):
    endpoint_config_class = DistributionTableEndpointConfig
    display_config_class = DistributionTableDisplayConfig
    title_config_class = DistributionTableTitleConfig
    button_config_class = DistributionTableButtonConfig

    def get_pandas_fields(self, request):
        if self.request.GET.get("group_by") != "INDUSTRY":
            fields = [
                pf.PKField(key="aggregate_field", label=""),
            ]
        else:
            fields = [
                pf.PKField(key="id", label="IDS"),
                pf.CharField(key="equity", label="Equity"),
            ]
            for level_rep in self.classification_levels_representation:
                fields.append(pf.CharField(key=level_rep, label=level_rep))
        fields.extend([pf.FloatField(key="weighting", label="Weight", precision=2, percent=True)])
        return pf.PandasFields(fields=tuple(fields))

    def get_date_filter(self):
        if date_str := self.request.GET.get("date", None):
            val_date = dt.datetime.strptime(date_str, "%Y-%m-%d").date()
        elif super().get_queryset().exists():
            val_date = AssetPosition.objects.latest("date").date
        else:
            val_date = dt.date.today()
        return val_date

    def get_queryset(self):
        val_date = self.get_date_filter()
        queryset = super().get_queryset().filter(date=val_date)
        return queryset

    def get_dataframe(self, request, queryset, **kwargs):
        group_by = self.request.GET.get("group_by", "COUNTRY")
        class_method = AssetPositionGroupBy.get_class_method_group_by(name=group_by)
        queryset_without_cash = queryset.exclude(underlying_instrument__is_cash=True)
        if group_by not in ["INDUSTRY", "CURRENCY"]:
            df = self.dataframe_groupby_with_class_method(qs=queryset_without_cash, class_method=class_method)
        elif group_by == "CURRENCY":
            df = self.dataframe_groupby_with_class_method(qs=queryset, class_method=class_method)
        else:  # group_by == "INDUSTRY"
            df = self._generate_classification_df(queryset_without_cash)

            if not df.empty:
                df.weighting /= df.weighting.sum()
                df = df.reset_index().rename(columns={**self.classification_columns_map, "index": "equity"})
                df["equity"] = df["equity"].map(
                    dict(Instrument.objects.filter(id__in=df["equity"]).values_list("id", "name_repr"))
                )
                for level in self.classification_levels_representation:
                    tmp = df.groupby(by=level).weighting.sum().astype(float).mul(100).round(1)
                    df = df.join(tmp, on=level, rsuffix=f"_{level}")
                    df[level] += " (" + df[f"weighting_{level}"].astype(str) + "%)"
                portfolio = Portfolio.objects.get(id=self.request.GET.get("portfolio"))
                if not PortfolioRole.is_analyst(self.request.user.profile, portfolio=portfolio):
                    df[["weighting", "equity"]] = None
                    df.drop_duplicates(inplace=True)
        return df

    def manipulate_dataframe(self, df):
        if not df.empty:
            df.sort_values(by="weighting", ascending=False, inplace=True)
            if df.weighting.sum() != 1:  # normalize
                df.weighting /= df.weighting.sum()
            df = df.reset_index(names="aggregate_field" if self.request.GET.get("group_by") != "INDUSTRY" else "id")
        return df



import pytest
from faker import Faker
from pandas.tseries.offsets import BDay
from wbfdm.factories import InstrumentFactory
from wbfdm.models import InstrumentPrice

from wbportfolio.models import (
    Adjustment,
    PortfolioInstrumentPreferredClassificationThroughModel,
)

fake = Faker()


@pytest.mark.django_db
class TestMergeInstrument:
    @pytest.fixture()
    def merged_instrument(self):
        return InstrumentFactory.create()

    @pytest.fixture()
    def main_instrument(self):
        return InstrumentFactory.create()

    def test_assets(
        self, main_instrument, merged_instrument, asset_position_factory, instrument_price_factory, weekday
    ):
        """
        Test if the asset positions attached to the merged instrument are forwarded to the main instrument as well as the underlying instrument price
        """
        p1 = instrument_price_factory.create(instrument=main_instrument, calculated=False, date=weekday)
        p2 = instrument_price_factory.create(instrument=merged_instrument, calculated=False, date=weekday)

        a1 = asset_position_factory.create(underlying_instrument=main_instrument, date=weekday)
        a2 = asset_position_factory.create(underlying_instrument=merged_instrument, date=weekday)

        assert a1.underlying_quote_price == p1
        assert a2.underlying_quote_price == p2
        main_instrument.merge(merged_instrument)
        a2.refresh_from_db()
        a1.refresh_from_db()

        assert a1.underlying_instrument == main_instrument  # Make sure this doesn't change
        assert a1.underlying_quote_price == p1  # Make sure this doesn't change
        assert a2.underlying_instrument == main_instrument
        assert a2.underlying_quote_price == p1

    def test_roles(self, main_instrument, merged_instrument, product_portfolio_role_factory):
        """
        Test if the role attached to the merged instrument are forwarded to the main instrument
        """
        role = product_portfolio_role_factory.create(instrument=merged_instrument)
        main_instrument.merge(merged_instrument)
        role.refresh_from_db()
        assert role.instrument == main_instrument

    def test_classifications(self, main_instrument, merged_instrument, portfolio_factory, classification_factory):
        """
        Test if preferred classification are forwarded to the main instrument
        """
        main_portfolio = portfolio_factory.create()
        merged_portfolio = portfolio_factory.create()

        PortfolioInstrumentPreferredClassificationThroughModel.objects.create(
            instrument=main_instrument, portfolio=main_portfolio, classification=classification_factory.create()
        )
        PortfolioInstrumentPreferredClassificationThroughModel.objects.create(
            instrument=merged_instrument, portfolio=merged_portfolio, classification=classification_factory.create()
        )
        main_instrument.merge(merged_instrument)

        assert set(main_instrument.preferred_portfolio_classifications.all()) == set(
            [main_portfolio, merged_portfolio]
        )

    def test_transactions(self, main_instrument, merged_instrument, trade_factory):
        """
        Test if the attached transactions to the merged instrument are forwarded to the main instrument
        """
        merged_t = trade_factory.create(underlying_instrument=merged_instrument)
        main_t = trade_factory.create(underlying_instrument=main_instrument)

        main_instrument.merge(merged_instrument)
        merged_t.refresh_from_db()
        main_t.refresh_from_db()

        merged_t.underlying_instrument = main_instrument
        main_t.underlying_instrument = main_instrument  # make sure this does not change

    def test_instrument_prices(self, main_instrument, merged_instrument, instrument_price_factory, weekday):
        """
        Test if the attached prices to the merged instrument are forwarded to the main instrument but does not overlaps them (if price already exists, the overlapping merged price are simply deleted)
        """
        main_p1 = instrument_price_factory.create(instrument=main_instrument, calculated=False, date=weekday)

        merged_p1 = instrument_price_factory.create(instrument=merged_instrument, calculated=False, date=weekday)
        merged_p2 = instrument_price_factory.create(
            instrument=merged_instrument, calculated=False, date=weekday + BDay(1)
        )

        assert main_instrument.valuations.count() == 1
        main_instrument.merge(merged_instrument)

        assert (
            main_instrument.valuations.get(date=weekday).net_value == main_p1.net_value
        )  # Check that the existing price was not overlaps by the merged instrument price at that day
        assert (
            main_instrument.valuations.get(date=weekday + BDay(1)).net_value == merged_p2.net_value
        )  # Check that the price not existing from merged instrument was appended to the main instrument serie
        with pytest.raises(InstrumentPrice.DoesNotExist):
            merged_p1.refresh_from_db()

    def test_adjustments(self, main_instrument, merged_instrument, adjustment_factory, weekday):
        """
        Test if the attached adjustments to the merged instrument are forwarded to the main instrument but does not overlaps them (if split already exists, the overlapping merged split are simply deleted)
        """
        main_s1 = adjustment_factory.create(instrument=main_instrument, date=weekday)

        merged_s1 = adjustment_factory.create(instrument=merged_instrument, date=weekday)
        merged_s2 = adjustment_factory.create(instrument=merged_instrument, date=weekday + BDay(1))
        merged_s2.refresh_from_db()  # We refresh from db to cast the decimal factor to its proper number of decimal places (kind of stupid but..)
        main_s1.refresh_from_db()

        assert main_instrument.pms_adjustments.count() == 1
        main_instrument.merge(merged_instrument)

        assert (
            main_instrument.pms_adjustments.get(date=weekday).factor == main_s1.factor
        )  # Check that the existing split was not overlaps by the merged instrument split at that day
        assert (
            main_instrument.pms_adjustments.get(date=weekday + BDay(1)).factor == merged_s2.factor
        )  # Check that the split not existing from merged instrument was appended to the main instrument serie
        with pytest.raises(Adjustment.DoesNotExist):
            merged_s1.refresh_from_db()
